from collections.abc import Collection
from dataclasses import dataclass

from elasticai.creator.vhdl.code import CodeFile, CodeModuleBase
from elasticai.creator.vhdl.code_files.fp_hard_sigmoid_file import FPHardSigmoidFile
from elasticai.creator.vhdl.number_representations import FixedPointFactory


@dataclass
class FPHardSigmoidModule(CodeModuleBase):
    layer_id: str
    fixed_point_factory: FixedPointFactory

    @property
    def name(self) -> str:
        return self.layer_id

    @property
    def files(self) -> Collection[CodeFile]:
        yield FPHardSigmoidFile(
            layer_id=self.layer_id,
            zero_threshold=self.fixed_point_factory(-3),
            one_threshold=self.fixed_point_factory(3),
            slope=self.fixed_point_factory(0.125),
            y_intercept=self.fixed_point_factory(0.5),
            fixed_point_factory=self.fixed_point_factory,
        )



from dataclasses import dataclass

import torch
from torch.nn.parameter import Parameter

from elasticai.creator.nn.arithmetics import Arithmetics, FloatArithmetics
from elasticai.creator.nn.linear import Linear
from elasticai.creator.tests.unit.tensor_test_case import TensorTestCase


def tensor(data: list) -> torch.Tensor:
    return torch.as_tensor(data, dtype=torch.float32)


def linear_base_with_fixed_params(
    in_features: int,
    out_features: int,
    bias: bool,
    arithmetics: Arithmetics = FloatArithmetics(),
) -> Linear:
    linear = Linear(
        in_features=in_features,
        out_features=out_features,
        arithmetics=arithmetics,
        bias=bias,
    )
    linear.weight = Parameter(torch.ones_like(linear.weight))
    if bias:
        linear.bias = Parameter(torch.ones_like(linear.bias))
    return linear


class AddThreeArithmetics(FloatArithmetics):
    def quantize(self, a: torch.Tensor) -> torch.Tensor:
        return a + 3


@dataclass
class FixedMatmulResultArithmetics(FloatArithmetics):
    value: list[float]

    def matmul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return torch.tensor(self.value)


@dataclass
class FixedAddResultArithmetics(FloatArithmetics):
    value: list[float]

    def add(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return torch.tensor(self.value)


class LinearTest(TensorTestCase):
    def test_with_bias(self) -> None:
        linear = linear_base_with_fixed_params(in_features=3, out_features=1, bias=True)

        actual = linear(tensor([1, 2, 3]))
        expected = [7.0]

        self.assertTensorEqual(expected, actual)

    def test_without_bias(self) -> None:
        linear = linear_base_with_fixed_params(
            in_features=3, out_features=1, bias=False
        )

        actual = linear(tensor([1, 2, 3]))
        expected = [6.0]

        self.assertTensorEqual(expected, actual)

    def test_different_quantization(self) -> None:
        linear = linear_base_with_fixed_params(
            in_features=3, out_features=1, bias=True, arithmetics=AddThreeArithmetics()
        )

        actual = linear(tensor([1, 2, 3]))
        expected = [28.0]

        self.assertTensorEqual(expected, actual)

    def test_result_is_3_with_fixed_matmul(self) -> None:
        expected = [3.0]
        linear = linear_base_with_fixed_params(
            in_features=3,
            out_features=1,
            bias=False,
            arithmetics=FixedMatmulResultArithmetics(expected),
        )

        actual = linear(tensor([1, 2, 3]))

        self.assertTensorEqual(expected, actual)

    def test_result_is_0_with_fixed_add(self) -> None:
        expected = [1.0]
        linear = linear_base_with_fixed_params(
            in_features=3,
            out_features=1,
            bias=True,
            arithmetics=FixedAddResultArithmetics(expected),
        )

        actual = linear(tensor([1, 2, 3]))

        self.assertTensorEqual(expected, actual)



import torch

from elasticai.creator.vhdl.number_representations import (
    FixedPoint,
    float_values_to_fixed_point,
)


def to_list(x: torch.Tensor) -> list:
    return x.detach().numpy().tolist()


def to_fixed_point(values: list[float], total_bits: int, frac_bits: int) -> list[int]:
    def to_fp(value: FixedPoint) -> int:
        return value.to_signed_int()

    return list(map(to_fp, float_values_to_fixed_point(values, total_bits, frac_bits)))


def from_fixed_point(values: list[int], total_bits: int, frac_bits: int) -> list[float]:
    def to_float(value: int) -> float:
        return float(FixedPoint.from_signed_int(value, total_bits, frac_bits))

    return list(map(to_float, values))



from dataclasses import dataclass
from typing import Iterator

from elasticai.creator.vhdl.code import CodeFile
from elasticai.creator.vhdl.language import Code
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.precomputed_scalar_function import Tanh


@dataclass
class TanhComponent(CodeFile):
    x: list[FixedPoint]

    @property
    def name(self) -> str:
        return f"tanh.vhd"

    def __iter__(self) -> Iterator[str]:
        yield from Tanh(x=self.x, component_name="tanh").code()

    def code(self) -> Code:
        return self



import torch

from elasticai.creator.nn.lstm import FixedPointLSTMWithHardActivations
from elasticai.creator.vhdl.translator.abstract.layers import LSTMModule


def build_fixed_point_lstm(
    layer: FixedPointLSTMWithHardActivations,
    layer_id: str,
    work_library_name: str,
) -> LSTMModule:
    def to_list(tensor: torch.Tensor) -> list:
        return tensor.detach().numpy().tolist()

    return LSTMModule(
        weights_ih=[to_list(layer.cell.linear_ih.weight)],
        weights_hh=[to_list(layer.cell.linear_hh.weight)],
        biases_ih=[to_list(layer.cell.linear_ih.bias)],
        biases_hh=[to_list(layer.cell.linear_hh.bias)],
        layer_id=layer_id,
        fixed_point_factory=layer.fixed_point_factory,
        work_library_name=work_library_name,
    )



from collections.abc import Collection, Iterable
from dataclasses import dataclass, field
from itertools import chain

import numpy as np

from elasticai.creator.vhdl.code import CodeFile, CodeModuleBase
from elasticai.creator.vhdl.code_files.fp_linear_1d_component import FPLinear1dFile
from elasticai.creator.vhdl.code_files.rom_component import RomFile
from elasticai.creator.vhdl.number_representations import FixedPoint, FixedPointFactory


@dataclass
class FPLinear1dModule(CodeModuleBase):
    layer_id: str
    weight: list[list[float]]
    bias: list[float]
    fixed_point_factory: FixedPointFactory
    work_library_name: str = field(default="work")

    @property
    def name(self) -> str:
        return self.layer_id

    @property
    def files(self) -> Collection[CodeFile]:
        def to_fp(values: Iterable[float]) -> list[FixedPoint]:
            return list(map(self.fixed_point_factory, values))

        out_features, in_features = np.shape(self.weight)

        yield FPLinear1dFile(
            layer_id=self.layer_id,
            in_feature_num=in_features,
            out_feature_num=out_features,
            fixed_point_factory=self.fixed_point_factory,
            work_library_name=self.work_library_name,
            resource_option="auto",
        )

        flat_weight = chain(*self.weight)

        name_suffix = f"_fp_linear_1d_{self.layer_id}"
        yield RomFile(
            rom_name="w_rom" + name_suffix,
            layer_id=self.layer_id,
            values=to_fp(flat_weight),
            resource_option="auto",
        )

        yield RomFile(
            rom_name="b_rom" + name_suffix,
            layer_id=self.layer_id,
            values=to_fp(self.bias),
            resource_option="auto",
        )



from unittest import TestCase

from torch import Tensor


def _to_list(x: Tensor) -> list:
    return x.detach().numpy().tolist()


class TensorTestCase(TestCase):
    def assertTensorEqual(self, expected: list | Tensor, actual: list | Tensor) -> None:
        expected = _to_list(expected) if isinstance(expected, Tensor) else expected
        actual = _to_list(actual) if isinstance(actual, Tensor) else actual
        self.assertEqual(expected, actual)



import unittest
from io import StringIO

from elasticai.creator.tests.integration.vhdl.vhd_file_reader import (
    VHDLFileReaderWithoutComments,
)


class VHDFileReaderWithoutCommentsTest(unittest.TestCase):
    def check(self, input: str, expected: str):
        io_dummy = StringIO(input)
        reader = VHDLFileReaderWithoutComments(io_dummy)
        expected_code = tuple(expected.splitlines())
        actual = tuple(reader)
        self.assertEqual(expected_code, actual)

    def test_ignores_blank_line(self):
        self.check(input="   \n vhdl file text", expected="vhdl file text")

    def test_ignore_line_comment(self):
        self.check("-- my comment\n my vhdl text", expected="my vhdl text")

    def test_ignore_trailing_comments(self):
        self.check("some vhdl code -- comment", expected="some vhdl code")

    def test_ignore_dashed_line(self):
        self.check("------\nsome text", expected="some text")

    def test_ignore_trailing_dashed_line(self):
        self.check("some text ------", expected="some text")

    def test_ignore_trailing_empty_comment(self):
        self.check("some text --", expected="some text")



import matplotlib.pyplot as plt
import torch
from torch.utils.data import DataLoader, Dataset, TensorDataset

from elasticai.creator.qat.layers import QLSTM, Identity, QLinear


def _x_values(value_range: tuple[float, float], sampling_rate: int) -> torch.Tensor:
    start, end = value_range
    return torch.linspace(
        start=start, end=end, steps=int((end - start) * sampling_rate)
    )


class SineDataset(Dataset):
    def __init__(
        self, sine_range: tuple[float, float], sampling_rate: int, window_size: int
    ) -> None:
        self.sine_range = sine_range
        self.sampling_rate = sampling_rate
        self.window_size = window_size

        x = _x_values(self.sine_range, self.sampling_rate)
        self._sinus = torch.sin(x)
        self._samples, self._labels = self._rolling_samples_with_labels(
            self._sinus, self.window_size
        )

    def __len__(self) -> int:
        return len(self._labels)

    def __getitem__(self, idx) -> tuple[torch.Tensor, torch.Tensor]:
        return self._samples[idx], self._labels[idx]

    @staticmethod
    def _rolling_samples_with_labels(
        data: torch.Tensor, window_size: int
    ) -> tuple[torch.Tensor, torch.Tensor]:
        samples, labels = [], []
        for i in range(window_size, len(data)):
            samples.append(data[i - window_size : i].view(-1, 1).tolist())
            labels.append(data[i].item())
        return (
            torch.tensor(samples, dtype=data.dtype),
            torch.tensor(labels, dtype=data.dtype),
        )

    @property
    def sinus_data(self):
        return self._sinus


class QLSTMModel(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.lstm = QLSTM(
            input_size=1,
            hidden_size=64,
            bias=True,
            batch_first=True,
            state_quantizer=Identity(),
            weight_quantizer=Identity(),
        )
        self.linear = QLinear(
            in_features=64,
            out_features=1,
            bias=True,
            quantizer=Identity(),
        )

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        _, (final_hidden_states, _) = self.lstm(inputs)
        x = self.linear(final_hidden_states.squeeze(0))
        return x.view(-1)


def split_dataset(dataset: Dataset, train_fraction: float) -> tuple[Dataset, Dataset]:
    num_train_samples = int(len(dataset) * train_fraction)
    ds_train = TensorDataset(*dataset[:num_train_samples])
    ds_test = TensorDataset(*dataset[num_train_samples:])
    return ds_train, ds_test


def train(
    model: torch.nn.Module,
    train_data: Dataset,
    test_data: Dataset,
    batch_size: int,
    learning_rate: float,
    epochs: int,
) -> None:
    dl_train = DataLoader(train_data, batch_size=batch_size, shuffle=True)
    dl_test = DataLoader(test_data, batch_size=batch_size, shuffle=True)

    loss_fn = torch.nn.MSELoss()
    optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)

    for epoch in range(epochs):
        losses = []
        for samples, labels in dl_train:
            predictions = model(samples)
            loss = loss_fn(predictions, labels)
            loss.backward()
            optimizer.step()
            optimizer.zero_grad()
            losses.append(loss.item())
        train_loss = sum(losses) / len(losses)

        with torch.no_grad():
            val_losses = map(lambda x: loss_fn(model(x[0]), x[1]), dl_test)
            val_loss = sum(val_losses) / len(dl_test)

        print(f"[epoch: {epoch + 1}/{epochs}] loss: {train_loss}; val_loss: {val_loss}")

    model.eval()


def plot_predicted_sine(
    model: torch.nn.Module,
    prediction_range: tuple[int, int],
    sampling_rate: int,
    window_size: int,
) -> None:
    pred_start, pred_end = prediction_range
    x_prelude = _x_values(
        (pred_start - window_size / sampling_rate, pred_start), sampling_rate
    )
    x_to_predict = _x_values(prediction_range, sampling_rate)
    y_prelude = torch.sin(x_prelude)
    y_target = torch.sin(x_to_predict)

    predictions: list[float] = y_prelude.tolist()
    for i in range(int((pred_end - pred_start) * sampling_rate)):
        inputs = torch.tensor(predictions[i:]).view(-1, 1)
        prediction = model(inputs)
        predictions.append(prediction[0].item())
    y_pred = torch.tensor(predictions[window_size:])

    plt.plot(x_prelude, y_prelude, "b-", label="prelude")
    plt.plot(x_to_predict, y_target, "g-", label="target")
    plt.plot(x_to_predict, y_pred, "r--", label="prediction")
    plt.legend()
    plt.show()


def main() -> None:
    ds = SineDataset(sine_range=(0, 200), sampling_rate=16, window_size=32)
    print("Number of samples:", len(ds))

    plt.plot(ds.sinus_data)
    plt.show()

    model = QLSTMModel()
    ds_train, ds_test = split_dataset(ds, train_fraction=0.8)

    train(
        model=model,
        train_data=ds_train,
        test_data=ds_test,
        batch_size=32,
        learning_rate=1e-3,
        epochs=10,
    )

    plot_predicted_sine(
        model=model,
        prediction_range=(400, 410),
        sampling_rate=16,
        window_size=32,
    )


if __name__ == "__main__":
    main()



from elasticai.creator.nn.relu import ReLU
from elasticai.creator.vhdl.number_representations import FixedPointFactory
from elasticai.creator.vhdl.translator.abstract.layers.fp_relu_module import (
    FPReLUModule,
)


def build_fp_relu(
    layer: ReLU, layer_id: str, fixed_point_factory: FixedPointFactory
) -> FPReLUModule:
    return FPReLUModule(layer_id=layer_id, fixed_point_factory=fixed_point_factory)



import unittest
from typing import Iterable

import torch

from elasticai.creator.vhdl.code import (
    Code,
    CodeFile,
    CodeFileBase,
    CodeModule,
    CodeModuleBase,
)
from elasticai.creator.vhdl.translator.build_function_mapping import (
    BuildFunctionMapping,
)
from elasticai.creator.vhdl.translator.pytorch import translator
from elasticai.creator.vhdl.translator.pytorch.build_function_mappings import (
    DEFAULT_BUILD_FUNCTION_MAPPING,
)


def fake_build_function(module: torch.nn.Module, layer_id: str) -> CodeModuleBase:
    return CodeModuleBase(
        name="module0",
        files=[
            CodeFileBase(name="component1", code=["1", "2", "3"]),
            CodeFileBase(name="component2", code=["4", "5", "6"]),
        ],
    )


def unpack_module_directories(
    modules: Iterable[CodeModule],
) -> list[tuple[str, list[tuple[str, Code]]]]:
    def unpack_code_file(code_file: CodeFile) -> tuple[str, Code]:
        return code_file.name, list(code_file.code())

    return [
        (module.name, list(map(unpack_code_file, module.files))) for module in modules
    ]


class TranslatorTest(unittest.TestCase):
    def setUp(self) -> None:
        self.build_mapping = BuildFunctionMapping(
            mapping={"torch.nn.modules.module.Module": fake_build_function}
        )

    def test_translate_model_empty_model(self) -> None:
        model = torch.nn.Sequential()
        generated_code = translator.translate_model(
            model, translation_args=dict(), build_function_mapping=self.build_mapping
        )
        self.assertEqual(len(list(generated_code)), 1)

    def test_translation_yields_one_build_mapping_result_for_first_layer(self) -> None:
        model = torch.nn.Sequential(torch.nn.Module())
        code_containers = translator.translate_model(
            model, translation_args=dict(), build_function_mapping=self.build_mapping
        )

        code = unpack_module_directories(code_containers)
        expected_code = (
            "0_Module",
            [
                ("component1", ["1", "2", "3"]),
                ("component2", ["4", "5", "6"]),
            ],
        )

        self.assertEqual(code[0], expected_code)

    @unittest.SkipTest
    def test_translate_model_correct_ordering_of_layers(self):
        class Model(torch.nn.Module):
            def __init__(self) -> None:
                super().__init__()
                self.lstm = torch.nn.LSTM(input_size=1, hidden_size=2)
                self.linear = torch.nn.Linear(in_features=2, out_features=1)

            def forward(self, x: torch.Tensor) -> torch.Tensor:
                return self.lstm(self.linear(x))

        model = Model()
        translated = list(
            translator.translate_model(
                model,
                build_function_mapping=DEFAULT_BUILD_FUNCTION_MAPPING,
            )
        )

        self.assertEqual(translated[0].name, "0_Linear")
        self.assertEqual(translated[1].name, "1_LSTM")



from abc import abstractmethod
from itertools import chain
from typing import Any, Callable, Collection, Iterable, Protocol, runtime_checkable

import torch
from torch.fx import Graph as fxGraph
from torch.fx import Tracer as fxTracer

from elasticai.creator.mlframework import Module
from elasticai.creator.vhdl.code import Code
from elasticai.creator.vhdl.hw_equivalent_layers.typing import HWEquivalentLayer


class Node(Protocol):
    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @property
    @abstractmethod
    def next(self) -> "Node":
        ...

    @property
    @abstractmethod
    def op(self) -> str:
        ...


@runtime_checkable
class HWEquivalentNode(Node, Protocol):
    @property
    @abstractmethod
    def hw_equivalent_layer(self) -> HWEquivalentLayer:
        ...


class HWBlockCollection(Protocol):
    @abstractmethod
    def signals(self, prefix: str) -> Code:
        ...

    @abstractmethod
    def instantiations(self, prefix: str) -> Code:
        ...


class HWEquivalentGraph(Protocol):
    @property
    @abstractmethod
    def hw_equivalent_nodes(self) -> Iterable[HWEquivalentNode]:
        ...

    @property
    @abstractmethod
    def nodes(self) -> Iterable[Node]:
        ...


class Tracer(Protocol):
    @abstractmethod
    def trace(self, model: Module) -> HWEquivalentGraph:
        ...


class HWEquivalentTracer(Tracer):
    def __init__(self):
        self._tracer = _HWEquivalentTracer()

    def trace(self, model: Module) -> HWEquivalentGraph:
        return self._tracer.trace(
            model,
        )


def create_hw_block_collection(graph: HWEquivalentGraph) -> HWBlockCollection:
    if not isinstance(graph, _HWEquivalentGraph):
        raise NotImplementedError(
            "hw block creation unsupported for {}".format(type(graph))
        )
    else:
        return graph


class _HWEquivalentGraph(HWEquivalentGraph, HWBlockCollection):
    """
        The HWEquivalentGraph is the result of tracing a compatible neural network `m`
    with the corresponding HWEquivalentTracer. It combines signal and
    portmaps for instantiation for all nodes linked to HWEquivalent submodules of `m`
    by making calls to these submodules.
    """

    def signals(self, prefix: str) -> Code:
        yield from chain.from_iterable(
            (
                node.hw_equivalent_layer.signal_definitions(f"{prefix}{node.name}")
                for node in self.hw_equivalent_nodes
            )
        )

    def _call_on_layers(self, method_name, prefix: str) -> Code:
        yield from chain.from_iterable(
            (
                getattr(node.hw_equivalent_layer, method_name)(f"{prefix}{node.name}")
                for node in self.hw_equivalent_nodes
            )
        )

    def instantiations(self, prefix: str) -> Code:
        yield from chain.from_iterable(
            (
                node.hw_equivalent_layer.instantiation(f"{prefix}{node.name}")
                for node in self.hw_equivalent_nodes
            )
        )

    def __init__(self, fx_graph: fxGraph):
        self._fx_graph = fx_graph

    @property
    def nodes(self):
        yield from self._fx_graph.nodes

    @property
    def hw_equivalent_nodes(self) -> Iterable[HWEquivalentNode]:
        yield from filter(lambda n: n.op == "call_module", self._fx_graph.nodes)

    @property
    def hw_equivalent_layers(self) -> Collection[HWEquivalentLayer]:
        layers = set()
        for node in self.hw_equivalent_nodes:
            if isinstance(node, HWEquivalentNode):
                layers.add(node.hw_equivalent_layer)
        return layers


class _HWEquivalentTracer(fxTracer):
    def is_leaf_module(self, m: torch.nn.Module, module_qualified_name: str) -> bool:
        return m.__module__.startswith("elasticai.creator.vhdl.hw_equivalent_layers")

    def trace(self, root, **kwargs) -> HWEquivalentGraph:
        graph = super().trace(root, **kwargs)
        return _HWEquivalentGraph(graph)

    def call_module(
        self,
        m: torch.nn.Module,
        forward: Callable[..., Any],
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> Any:
        proxy = super().call_module(m, forward, args, kwargs)
        proxy.node.hw_equivalent_layer = m
        return proxy



from typing import Iterable
from unittest import TestCase

from elasticai.creator.resource_utils import get_file
from elasticai.creator.tests.integration.vhdl.code_test_case import CodeTestCase
from elasticai.creator.tests.integration.vhdl.vhd_file_reader import (
    VHDLFileReaderWithoutComments,
)
from elasticai.creator.tests.text_parsing import extract_section
from elasticai.creator.vhdl.code import Code
from elasticai.creator.vhdl.hw_equivalent_layers.layers import (
    FixedPointHardSigmoid,
    FixedPointLinear,
    RootModule,
)
from elasticai.creator.vhdl.number_representations import ClippedFixedPoint

"""
Tests:
    - Names of generated CodeModules are converted torch module_hierarchy names.
    - For every leaf module in our torch module we generate a CodeModule
    - For every node in our graph from tracing torch module we instantiate a CodeModule of the correct name in our network.vhd
"""


class FirstModel(RootModule):
    def __init__(self):
        super().__init__()
        self.data_width = 16
        fp_factory = ClippedFixedPoint.get_factory(
            total_bits=self.data_width, frac_bits=8
        )
        self.fp_linear = FixedPointLinear(
            in_features=2,
            out_features=2,
            fixed_point_factory=fp_factory,
            data_width=self.data_width,
        )
        self.fp_hard_sigmoid = FixedPointHardSigmoid(data_width=self.data_width)

    def forward(self, x):
        return self.fp_hard_sigmoid(self.fp_linear(x))


class GenerateNetworkRootFileFromDifferentModelVersions(TestCase):
    def setUp(self):
        self.model = FirstModel()

        with get_file(
            "elasticai.creator.tests.integration.vhdl", "expected_network.vhd"
        ) as f:
            self.expected_code = VHDLFileReaderWithoutComments(f).as_list()

    def get_generated_portmap_signal_line(self, signal_id, value) -> str:
        self.model.elasticai_tags.update({f"{signal_id}_width": value})
        code = CodeTestCase.unified_vhdl_from_module(self.model.translate())
        portmap = extract_section(begin="port (", end=");", lines=code)[0]
        actual = next(filter(lambda line: line.startswith(f"{signal_id}:"), portmap))
        return actual

    def check_y_address_width(self, value: int):
        actual = self.get_generated_portmap_signal_line("y_address", value)
        self.assertEqual(f"y_address: in std_logic_vector({value}-1 downto 0);", actual)

    def check_x_data_width(self, value: int):
        actual = self.get_generated_portmap_signal_line("x", value)
        self.assertEqual(f"x: in std_logic_vector({value}-1 downto 0);", actual)

    def test_portmap_output_addr_width_is_4(
        self,
    ):
        self.check_y_address_width(4)

    def test_portmap_output_addr_width_is_8(self):
        self.check_y_address_width(8)

    def test_x_data_width_is_4(self):
        self.check_x_data_width(4)

    def test_y_data_width_is_8(self):
        actual = self.get_generated_portmap_signal_line("y", 8)
        self.assertEqual("y: out std_logic_vector(8-1 downto 0);", actual)

    def test_x_address_width_is_16(self):
        actual = self.get_generated_portmap_signal_line("x_address", 16)
        self.assertEqual("x_address: out std_logic_vector(16-1 downto 0);", actual)


class GeneratedNetworkVHDMatchesTargetForSingleModelVersion(CodeTestCase):
    """
    Tests:
      - each of the port maps is generated correctly
      - signals for each layer are connected correctly
    """

    def setUp(self):
        self.model = FirstModel()
        self.model.elasticai_tags.update(
            {
                "x_address_width": 1,
                "y_address_width": 1,
                "y_width": 16,
                "x_width": 16,
            }
        )
        self.actual_code = CodeTestCase.unified_vhdl_from_module(self.model.translate())

    def test_port_def_matches_target(self):
        expected_port_def = CodeTestCase.code_section_from_string(
            """
        enable: in std_logic;
        clock: in std_logic;
        x_address: out std_logic_vector(1-1 downto 0);
        y_address: in std_logic_vector(1-1 downto 0);
        x: in std_logic_vector(16-1 downto 0);
        y: out std_logic_vector(16-1 downto 0);
        done: out std_logic
        """
        )
        actual = extract_section(begin="port (", end=");", lines=self.actual_code)
        self.check_lines_are_equal_ignoring_order(expected_port_def, actual)

    def test_signal_defs_match_target(self):
        actual_code = self.actual_code
        expected_signal_defs = CodeTestCase.code_section_from_string(
            """
    -- fp_linear
    signal fp_linear_enable : std_logic := '0';
    signal fp_linear_clock : std_logic := '0';
    signal fp_linear_done : std_logic := '0';
    signal fp_linear_x : std_logic_vector(15 downto 0);
    signal fp_linear_y : std_logic_vector(15 downto 0);
    signal fp_linear_x_address : std_logic_vector(0 downto 0);
    signal fp_linear_y_address : std_logic_vector(0 downto 0);

    -- fp_hard_sigmoid
    signal fp_hard_sigmoid_enable : std_logic := '0';
    signal fp_hard_sigmoid_clock : std_logic := '0';
    signal fp_hard_sigmoid_x : std_logic_vector(15 downto 0);
    signal fp_hard_sigmoid_y : std_logic_vector(15 downto 0);
            """
        )
        actual_sections = extract_section(
            begin="architecture rtl of fp_network is", end="begin", lines=actual_code
        )
        self.check_lines_are_equal_ignoring_order(expected_signal_defs, actual_sections)

    def test_fp_linear_portmap_is_generated(self):
        self.check_portmaps(
            expected="""
        enable => fp_linear_enable,
        clock => fp_linear_clock,

        x => fp_linear_x,
        y => fp_linear_y,
        x_address => fp_linear_x_address,
        y_address => fp_linear_y_address,
        done => fp_linear_done
        """,
            portmap_start="fp_linear : entity work.fp_linear(rtl)",
        )

    @staticmethod
    def strip_comma_from_portmaps(portmaps: Iterable[Code]) -> Iterable[Code]:
        for portmap in portmaps:
            yield (line.rstrip(",") for line in portmap)

    def test_all_but_last_portmap_line_have_trailing_comma(self):
        portmaps = extract_section(begin="port map(", end=");", lines=self.actual_code)
        for counter, portmap in enumerate(portmaps):
            with self.subTest(f"portmap number {counter}"):
                with self.subTest("lines end with comma"):
                    self.assertTrue(all((line.endswith(",") for line in portmap[:-1])))
                with self.subTest("last line has no comma"):
                    self.assertFalse(portmap[-1].endswith(","))

    def check_portmaps(self, expected: str, portmap_start: str):
        expected_portmap: Iterable[Code] = CodeTestCase.code_section_from_string(
            expected
        )
        actual: Iterable[Code] = extract_section(
            begin=[portmap_start, "port map("], end=");", lines=self.actual_code
        )
        expected_portmap = self.strip_comma_from_portmaps(expected_portmap)
        actual = self.strip_comma_from_portmaps(actual)
        self.check_lines_are_equal_ignoring_order(expected_portmap, actual)

    def test_fp_hard_sigmoid_portmap_is_generated(self):
        self.check_portmaps(
            expected="""
        enable => fp_hard_sigmoid_enable,
        clock => fp_hard_sigmoid_clock,
        x => fp_hard_sigmoid_x,
        y => fp_hard_sigmoid_y""",
            portmap_start="fp_hard_sigmoid : entity work.fp_hard_sigmoid(rtl)",
        )


class SignalConnectionsTest(CodeTestCase):
    def setUp(self) -> None:
        self.model = FirstModel()
        self.model.elasticai_tags.update(
            {
                "x_address_width": 1,
                "y_address_width": 1,
                "y_width": 16,
                "x_width": 16,
            }
        )
        code = CodeTestCase.unified_vhdl_from_module(self.model.translate())
        self.actual_connections: Code = extract_section(
            begin="begin",
            end="fp_linear : entity work.fp_linear(rtl)",
            lines=code,
        )[0]

    def test_x_is_connected_to_fp_linear_x(self):
        self.check_contains_all_expected_lines(
            expected=["fp_linear_x <= x;"], actual=self.actual_connections
        )



import unittest
from functools import partial

from elasticai.creator.vhdl.hw_equivalent_layers.hw_blocks import (
    BaseHWBlock,
    BufferedBaseHWBlock,
)


class NetworkHWComponentTest(unittest.TestCase):
    """
    Tests:
    - signals are uniquely named
    - signal names can be traced back to their emitting graph nodes and/or layers
    - all layers from model are instantiated
    - layer names correspond to signal and folder names
        -> node names need to be propagated to each module as well as network component
    - constants are correctly generated for
        - data width per layer connection
        - address width per layer connection

    Notes:
     - network needs to know in (x) and out (y) address widths for layers
     - network needs graph and component names for each graph node
     - network needs to know data width

    Questions for Chao:
        - Answer: address_width == number of bits to count number of inputs
            - max(1, math.ceil(math.log2(num_items)))
    """


class HWBlockInstantiationTest(unittest.TestCase):
    def test_creates_instance(self):
        block = BaseHWBlock(x_width=0, y_width=0)
        lines = block.instantiation("my_comp")
        first_line = list(lines)[0]
        expected = "my_comp : entity work.my_comp(rtl)"
        self.assertEqual(expected, first_line)

    def test_opens_portmap(self):
        block = BaseHWBlock(x_width=0, y_width=0)
        lines = block.instantiation("my_comp")
        second_line = list(lines)[1]
        expected = "port map("
        self.assertEqual(expected, second_line)

    def test_assigns_all_ports(self):
        block = BaseHWBlock(x_width=0, y_width=0)
        lines = partial(block.instantiation, "my_comp")
        for port in ("enable", "clock", "x", "y"):
            with self.subTest(port):
                self.assertTrue(
                    any((line.startswith(f"{port} => ") for line in lines())),
                    f"expected '{port} =>',\nfound: {list(lines())}",
                )

    def test_ports_are_connected_to_correctly_named_signals(self):
        block = BaseHWBlock(x_width=0, y_width=0)
        lines = block.instantiation("some_other_name")
        port_to_signal_connections = list(lines)[2:-1]
        for line in port_to_signal_connections:
            with self.subTest(line):
                line = line.strip(",")
                port = line.split(" => ")[0]
                signal = line.split(" => ")[1]
                self.assertEqual(signal, f"some_other_name_{port}")


class HWBlockSignals(unittest.TestCase):
    def constructor(self, name, data_width, *args):
        return BaseHWBlock(x_width=data_width, y_width=data_width).signal_definitions(
            name
        )

    def test_clock_signal_is_generated(self):
        signals = self.constructor("other_name", 2)
        self.assertTrue("signal other_name_clock : std_logic := '0';" in list(signals))

    def test_component_name_is_generated(self):
        signals = self.constructor("other_name", 2)
        self.assertTrue("signal other_name_enable : std_logic := '0';" in list(signals))

    def check_signals_contain_logic_vector(self, prefix, signals, suffix, width):
        expected = f"signal {prefix}_{suffix} : std_logic_vector({width - 1} downto 0);"
        signals = list(signals)
        self.assertTrue(
            expected in signals,
            msg=f"expected: {expected}\nfound: {signals}",
        )

    def test_generates_correct_number_of_lines(self):
        signals = self.constructor("other_name", 2)
        self.assertEqual(4, len(tuple(signals)))


class HWBlockSignals(HWBlockSignals):
    def constructor(self, name, data_width, *args):
        x_address_width = 1
        y_address_width = 1
        y_data_width = data_width
        x_data_width = data_width
        if len(args) == 2:
            x_address_width = args[0]
            y_address_width = args[1]
        return BufferedBaseHWBlock(
            x_width=x_data_width,
            y_width=y_data_width,
            x_address_width=x_address_width,
            y_address_width=y_address_width,
        ).signal_definitions(name)

    def test_5_downto_0_generated_for_x_addr_width_6(self):
        signals = self.constructor("other_name", 4, 6, 2)
        self.check_signals_contain_logic_vector("other_name", signals, "x_address", 6)

    def test_8_downto_0_generated_for_y_addr_width_9(self):
        signals = self.constructor("other_name", 4, 6, 9)
        self.check_signals_contain_logic_vector("other_name", signals, "y_address", 9)

    def test_generates_correct_number_of_lines(self):
        signals = self.constructor("other_name", 2, 2, 4)
        self.assertEqual(7, len(tuple(signals)))



import unittest

import torch
from torch.nn import Conv1d
from torch.nn.utils.parametrize import register_parametrization

from elasticai.creator.nn.layers import Binarize


class ParametrizeConvolutionTest(unittest.TestCase):
    def setUp(self) -> None:
        seed = 42
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)

    def test_parametrize_conv1d(self) -> None:
        layer = Conv1d(in_channels=2, out_channels=3, kernel_size=(1,), bias=False)

        register_parametrization(layer, "weight", Binarize())

        inputs = torch.tensor([[[1], [1]]], dtype=torch.float32)
        outputs = layer(inputs).detach().numpy().tolist()

        self.assertEqual(type(layer.parametrizations["weight"][0]), Binarize)
        self.assertEqual(outputs, [[[2.0], [0.0], [0.0]]])



import unittest

import torch

from elasticai.creator.nn.layers import (
    QLSTM,
    Binarize,
    ChannelShuffle,
    Identity,
    QConv1d,
    QConv2d,
    QLinear,
    QLSTMCell,
    QuantizeTwoBit,
    ResidualQuantization,
    Ternarize,
)


def round_tensor(
    *input_tensor: torch.Tensor, decimals: int
) -> tuple[torch.Tensor, ...]:
    return tuple(
        tensor.to(float).detach().apply_(lambda x: round(x, decimals))
        for tensor in input_tensor
    )


class TernarizeTest(unittest.TestCase):
    def test_quantize_different_values(self) -> None:
        layer = Ternarize()
        out = layer(torch.tensor([1.0, -1.0, 0.0]))
        self.assertEqual(out.tolist(), [1.0, -1.0, 0.0])

    def test_does_not_require_grad_by_default(self) -> None:
        layer = Ternarize()
        out = layer(torch.tensor([2.0]))
        self.assertEqual(out.requires_grad, False)

    def test_widening_is_trainable_when_trainable_equals_true(self) -> None:
        layer = Ternarize(zero_window_width=0.5, trainable=True)
        out = layer(torch.tensor([2.0]))
        self.assertEqual(out.requires_grad, True)


class ResidualQuantizationTest(unittest.TestCase):
    def test_quantize_different_values(self) -> None:
        layer = ResidualQuantization()
        out = layer(torch.tensor([[1.0, -1.0, 0.0]]))
        self.assertEqual(out.tolist(), [[1.0, -1.0, 1.0, 1.0, -1.0, -1.0]])

    def test_backward_pass_requires_grad(self) -> None:
        layer = ResidualQuantization()
        out = layer(torch.tensor([[2.0]]))
        self.assertEqual(out.requires_grad, True)


class QuantizeTwoBitTest(unittest.TestCase):
    def test_call_with_0_3_factor_and_2_as_input(self) -> None:
        layer = QuantizeTwoBit(0.3)
        out = layer(torch.tensor([[2.0]]))
        self.assertEqual(out.tolist(), [[1, 1.0]])

    def test_factors_are_trainable(self) -> None:
        layer = QuantizeTwoBit(0.3)
        out = layer(torch.tensor([[2.0]]))
        self.assertEqual(out.requires_grad, True)


class QLinearTest(unittest.TestCase):
    def test_binarized_call_with_bias(self) -> None:
        layer = QLinear(in_features=1, out_features=2, quantizer=Binarize())
        layer.weight = torch.tensor([[2.0], [-2.0]])
        layer.bias = torch.tensor([[3.0, -3.0]])
        input = torch.tensor([[3.0]])
        out = layer(input)
        self.assertEqual(out.tolist(), [[4.0, -4.0]])

    def test_backward_pass_changes_weight_gradient(self) -> None:
        layer = QLinear(in_features=1, out_features=2, quantizer=Binarize())
        layer.weight = torch.tensor([[0.6], [0.7]])
        old_grad = layer.parametrizations.weight.original.grad
        layer.bias = torch.tensor([[3.0, -3.0]])
        input = torch.tensor([[3.0]])
        out = layer(input)
        loss = (out - torch.tensor([-1.0])).sum()
        loss.backward()
        new_weight_grad = layer.parametrizations.weight.original.grad
        self.assertNotEqual(new_weight_grad[0], old_grad)
        self.assertNotEqual(new_weight_grad[1], old_grad)

    def test_throw_error_if_quantizer_is_none(self) -> None:
        with self.assertRaises(TypeError):
            _ = QLinear(in_features=1, out_features=1, quantizer=None)


class QConv1dTest(unittest.TestCase):
    def test_binarized_call_with_bias(self) -> None:
        layer = QConv1d(
            in_channels=1, out_channels=1, kernel_size=(2,), quantizer=Binarize()
        )
        layer.weight = torch.zeros(1, 1, 2)
        layer.bias = torch.tensor([3.0])
        input = torch.tensor([[[3.0, 2.0]]])
        out = layer(input)
        self.assertEqual(out.tolist(), [[[6.0]]])

    def test_backward_pass_changes_weight_gradient(self) -> None:
        layer = QConv1d(
            in_channels=1, out_channels=1, kernel_size=(2,), quantizer=Binarize()
        )
        layer.weight = torch.zeros(1, 1, 2)
        layer.bias = torch.tensor([3.0])
        old_grad = layer.parametrizations.weight.original.grad
        input = torch.tensor([[[3.0, 2.0]]])
        out = layer(input)
        loss = (out - torch.tensor([-1.0])).sum()
        loss.backward()
        new_weight_grad = layer.parametrizations.weight.original.grad
        self.assertNotEqual(new_weight_grad[0], old_grad)

    def test_throw_error_if_quantizer_is_none(self) -> None:
        with self.assertRaises(TypeError):
            _ = QConv1d(in_channels=1, out_channels=1, kernel_size=2, quantizer=None)


class QConv2dTest(unittest.TestCase):
    def test_binarized_call_with_bias(self) -> None:
        layer = QConv2d(
            in_channels=1, out_channels=1, kernel_size=2, quantizer=Binarize()
        )
        layer.weight = torch.zeros(1, 1, 2, 2)
        layer.bias = torch.tensor([3.0])
        input = torch.tensor([[[[3.0, 2.0], [3.0, 2.0]]]])
        out = layer(input)
        self.assertEqual(out.tolist(), [[[[11.0]]]])

    def test_backward_pass_changes_weight_gradient(self) -> None:
        layer = QConv2d(
            in_channels=1, out_channels=1, kernel_size=2, quantizer=Binarize()
        )
        layer.weight = torch.zeros(1, 1, 2, 2)
        layer.bias = torch.tensor([3.0])
        input = torch.tensor([[[[3.0, 2.0], [3.0, 2.0]]]])
        old_grad = layer.parametrizations.weight.original.grad
        out = layer(input)
        loss = (out - torch.tensor([-1.0])).sum()
        loss.backward()
        new_weight_grad = layer.parametrizations.weight.original.grad
        self.assertNotEqual(new_weight_grad[0], old_grad)

    def test_throw_error_if_quantizer_is_none(self) -> None:
        with self.assertRaises(TypeError):
            _ = QConv2d(in_channels=1, out_channels=1, kernel_size=2, quantizer=None)


class ChannelShuffleTest(unittest.TestCase):
    def test_input_and_output_same_with_1_group(self) -> None:
        layer = ChannelShuffle(groups=1)
        input = torch.rand((2, 2, 3, 2))
        output = layer(input)
        self.assertEqual(input.tolist(), output.tolist())

    def test_input_and_output_with_2_groups(self) -> None:
        layer = ChannelShuffle(groups=2)
        input = torch.tensor([[[1], [2], [3], [4]]])
        output = layer(input)
        expected = torch.tensor([[[1], [3], [2], [4]]])
        self.assertEqual(output.tolist(), expected.tolist())


class QLSTMCellTest(unittest.TestCase):
    def test_full_res_qlstm_cell_equal_pytorch_lstm_cell_without_bias(self) -> None:
        lstm_cell = torch.nn.LSTMCell(input_size=3, hidden_size=5, bias=False)
        qlstm_cell = QLSTMCell(
            input_size=3,
            hidden_size=5,
            state_quantizer=Identity(),
            weight_quantizer=Identity(),
            bias=False,
        )

        qlstm_cell.weight_ih = lstm_cell.weight_ih
        qlstm_cell.weight_hh = lstm_cell.weight_hh
        qlstm_cell.bias_ih = lstm_cell.bias_ih
        qlstm_cell.bias_hh = lstm_cell.bias_hh

        inp = torch.rand(1, 3)
        lstm_h1, lstm_c1 = lstm_cell(inp)
        qlstm_h1, qlstm_c1 = qlstm_cell(inp)

        lstm_h1, lstm_c1 = round_tensor(lstm_h1, lstm_c1, decimals=4)
        qlstm_h1, qlstm_c1 = round_tensor(qlstm_h1, qlstm_c1, decimals=4)

        self.assertEqual(lstm_h1.tolist(), qlstm_h1.tolist())
        self.assertEqual(lstm_c1.tolist(), qlstm_c1.tolist())

    def test_full_res_qlstm_cell_equal_pytorch_lstm_cell_with_bias(self) -> None:
        lstm_cell = torch.nn.LSTMCell(input_size=3, hidden_size=5, bias=True)
        qlstm_cell = QLSTMCell(
            input_size=3,
            hidden_size=5,
            state_quantizer=Identity(),
            weight_quantizer=Identity(),
        )

        qlstm_cell.weight_ih = lstm_cell.weight_ih
        qlstm_cell.weight_hh = lstm_cell.weight_hh
        qlstm_cell.bias_ih = lstm_cell.bias_ih
        qlstm_cell.bias_hh = lstm_cell.bias_hh

        inp = torch.rand(1, 3)
        lstm_h1, lstm_c1 = lstm_cell(inp)
        qlstm_h1, qlstm_c1 = qlstm_cell(inp)

        lstm_h1, lstm_c1 = round_tensor(lstm_h1, lstm_c1, decimals=4)
        qlstm_h1, qlstm_c1 = round_tensor(qlstm_h1, qlstm_c1, decimals=4)

        self.assertEqual(lstm_h1.tolist(), qlstm_h1.tolist())
        self.assertEqual(lstm_c1.tolist(), qlstm_c1.tolist())

    def test_binarized_state_and_weight(self) -> None:
        cell = QLSTMCell(
            input_size=1,
            hidden_size=1,
            state_quantizer=Binarize(),
            weight_quantizer=Binarize(),
        )

        cell.weight_ih = torch.nn.Parameter(torch.ones_like(cell.weight_ih))
        cell.weight_hh = torch.nn.Parameter(torch.ones_like(cell.weight_hh) * (-1))
        cell.bias_ih = torch.nn.Parameter(torch.ones_like(cell.bias_ih))
        cell.bias_hh = torch.nn.Parameter(torch.ones_like(cell.bias_hh) * (-1))

        inp = torch.as_tensor([[1.0]])
        actual_h1, actual_c1 = cell(inp)
        actual_h1, actual_c1 = round_tensor(actual_h1, actual_c1, decimals=4)
        target_h1, target_c1 = [[0.2311]], [[0.5]]

        self.assertEqual(target_h1, actual_h1.tolist())
        self.assertEqual(target_c1, actual_c1.tolist())

    def test_output_shape_on_flat_input_data(self) -> None:
        input_size = 2
        hidden_size = 4
        cell = QLSTMCell(
            input_size=input_size,
            hidden_size=hidden_size,
            state_quantizer=Identity(),
            weight_quantizer=Identity(),
        )
        inputs = torch.rand(input_size)
        h1, c1 = cell(inputs)
        self.assertEqual(h1.shape, (hidden_size,))
        self.assertEqual(c1.shape, (hidden_size,))

    def test_output_shape_on_batched_input_data(self) -> None:
        input_size = 2
        hidden_size = 4
        batch_size = 3
        cell = QLSTMCell(
            input_size=input_size,
            hidden_size=hidden_size,
            state_quantizer=Identity(),
            weight_quantizer=Identity(),
        )
        inputs = torch.rand(batch_size, input_size)
        h1, c1 = cell(inputs)
        self.assertEqual(h1.shape, (batch_size, hidden_size))
        self.assertEqual(c1.shape, (batch_size, hidden_size))


class QLSTMTest(unittest.TestCase):
    def test_output_shape_on_flat_input_data(self) -> None:
        input_size, hidden_size, sequence_len = 2, 4, 3
        lstm = QLSTM(
            input_size,
            hidden_size,
            state_quantizer=Identity(),
            weight_quantizer=Identity(),
        )
        inputs = torch.rand(sequence_len, input_size)
        h0, c0 = torch.rand(1, hidden_size), torch.rand(1, hidden_size)
        output, (h1, c1) = lstm(inputs, (h0, c0))
        self.assertEqual(output.shape, (sequence_len, hidden_size))
        self.assertEqual(h1.shape, (1, hidden_size))
        self.assertEqual(c1.shape, (1, hidden_size))

    def test_output_shape_on_batched_input_data(self) -> None:
        input_size, hidden_size, sequence_len = 2, 4, 3
        batch_size = 5
        lstm = QLSTM(
            input_size,
            hidden_size,
            state_quantizer=Identity(),
            weight_quantizer=Identity(),
        )
        inputs = torch.rand(sequence_len, batch_size, input_size)
        h0 = torch.rand(1, batch_size, hidden_size)
        c0 = torch.rand(1, batch_size, hidden_size)
        output, (h1, c1) = lstm(inputs, (h0, c0))
        self.assertEqual(output.shape, (sequence_len, batch_size, hidden_size))
        self.assertEqual(h1.shape, (1, batch_size, hidden_size))
        self.assertEqual(c1.shape, (1, batch_size, hidden_size))

    def test_output_shape_on_batched_input_data_batch_first(self) -> None:
        input_size, hidden_size, sequence_len = 2, 4, 3
        batch_size = 5
        lstm = QLSTM(
            input_size,
            hidden_size,
            batch_first=True,
            state_quantizer=Identity(),
            weight_quantizer=Identity(),
        )
        inputs = torch.rand(batch_size, sequence_len, input_size)
        h0 = torch.rand(1, batch_size, hidden_size)
        c0 = torch.rand(1, batch_size, hidden_size)
        output, (h1, c1) = lstm(inputs, (h0, c0))
        self.assertEqual(output.shape, (batch_size, sequence_len, hidden_size))
        self.assertEqual(h1.shape, (1, batch_size, hidden_size))
        self.assertEqual(c1.shape, (1, batch_size, hidden_size))

    def test_forward_without_explicit_given_state(self) -> None:
        pt_lstm = torch.nn.LSTM(input_size=1, hidden_size=2, bias=True)
        qlstm = QLSTM(
            input_size=1,
            hidden_size=2,
            state_quantizer=Identity(),
            weight_quantizer=Identity(),
        )

        qlstm.cell.weight_ih = pt_lstm.weight_ih_l0
        qlstm.cell.weight_hh = pt_lstm.weight_hh_l0
        qlstm.cell.bias_ih = pt_lstm.bias_ih_l0
        qlstm.cell.bias_hh = pt_lstm.bias_hh_l0

        inputs = torch.tensor([[1.0], [1.0]])
        pt_outputs, (pt_h, pt_c) = pt_lstm(inputs)
        q_outputs, (q_h, q_c) = qlstm(inputs)

        pt_outputs, pt_h, pt_c = round_tensor(pt_outputs, pt_h, pt_c, decimals=4)
        q_outputs, q_h, q_c = round_tensor(q_outputs, q_h, q_c, decimals=4)

        self.assertEqual(pt_outputs.tolist(), q_outputs.tolist())
        self.assertEqual(pt_h.tolist(), q_h.tolist())
        self.assertEqual(pt_c.tolist(), q_c.tolist())

    def test_forward_with_input_and_state(self) -> None:
        pt_lstm = torch.nn.LSTM(input_size=1, hidden_size=2, bias=True)
        qlstm = QLSTM(
            input_size=1,
            hidden_size=2,
            state_quantizer=Identity(),
            weight_quantizer=Identity(),
        )

        qlstm.cell.weight_ih = pt_lstm.weight_ih_l0
        qlstm.cell.weight_hh = pt_lstm.weight_hh_l0
        qlstm.cell.bias_ih = pt_lstm.bias_ih_l0
        qlstm.cell.bias_hh = pt_lstm.bias_hh_l0

        inputs = torch.tensor([[1.0], [1.0]])
        state = (torch.tensor([[5.0, 5.0]]), torch.tensor([[-1.0, -1.0]]))
        pt_outputs, (pt_h, pt_c) = pt_lstm(inputs, state)
        q_outputs, (q_h, q_c) = qlstm(inputs, state)

        pt_outputs, pt_h, pt_c = round_tensor(pt_outputs, pt_h, pt_c, decimals=4)
        q_outputs, q_h, q_c = round_tensor(q_outputs, q_h, q_c, decimals=4)

        self.assertEqual(pt_outputs.tolist(), q_outputs.tolist())
        self.assertEqual(pt_h.tolist(), q_h.tolist())
        self.assertEqual(pt_c.tolist(), q_c.tolist())

    def test_forward_flat_inputs_with_batch_first(self) -> None:
        pt_lstm = torch.nn.LSTM(input_size=1, hidden_size=2, batch_first=True)
        qlstm = QLSTM(
            input_size=1,
            hidden_size=2,
            batch_first=True,
            state_quantizer=Identity(),
            weight_quantizer=Identity(),
        )

        qlstm.cell.weight_ih = pt_lstm.weight_ih_l0
        qlstm.cell.weight_hh = pt_lstm.weight_hh_l0
        qlstm.cell.bias_ih = pt_lstm.bias_ih_l0
        qlstm.cell.bias_hh = pt_lstm.bias_hh_l0

        inputs = torch.tensor([[1.0], [1.0], [1.0]])
        state = (torch.tensor([[5.0, 5.0]]), torch.tensor([[-1.0, -1.0]]))
        pt_outputs, (pt_h, pt_c) = pt_lstm(inputs, state)
        q_outputs, (q_h, q_c) = qlstm(inputs, state)

        pt_outputs, pt_h, pt_c = round_tensor(pt_outputs, pt_h, pt_c, decimals=4)
        q_outputs, q_h, q_c = round_tensor(q_outputs, q_h, q_c, decimals=4)

        self.assertEqual(pt_outputs.tolist(), q_outputs.tolist())
        self.assertEqual(pt_h.tolist(), q_h.tolist())
        self.assertEqual(pt_c.tolist(), q_c.tolist())



from elasticai.creator.vhdl.translator.abstract.layers.fp_hard_sigmoid_module import (
    FPHardSigmoidModule,
)
from elasticai.creator.vhdl.translator.abstract.layers.fp_linear_1d_module import (
    FPLinear1dModule,
)
from elasticai.creator.vhdl.translator.abstract.layers.fp_relu_module import (
    FPReLUModule,
)
from elasticai.creator.vhdl.translator.abstract.layers.lstm_module import LSTMModule



"""
The module contains classes and functions for generating vhdl code similar to the language module
This module includes CodeGenerator that are only used by the vhdl testbenches
"""
from abc import ABC, abstractmethod
from typing import Iterator

from elasticai.creator.vhdl.language import Code


class TestBenchBase(ABC):
    simulation_start_msg = 'report "======Simulation Start======" severity Note'
    simulation_end_msgs = (
        'report "======Simulation Success======" severity Note',
        'report "Please check the output message." severity Note',
        "wait",
    )

    @abstractmethod
    def _body(self) -> Iterator[str]:
        ...

    def code(self) -> Code:
        def generator():
            yield self.simulation_start_msg
            yield from self._body()
            yield from self.simulation_end_msgs

        return generator()



import math
from typing import Callable

from elasticai.creator.vhdl.number_representations import (
    FixedPoint,
    infer_total_and_frac_bits,
)


def pad_with_zeros(numbers: list[FixedPoint], target_length: int) -> list[FixedPoint]:
    zero = FixedPoint(0, *infer_total_and_frac_bits(numbers))
    return numbers + [zero] * (target_length - len(numbers))


def derive_fixed_point_params_from_factory(
    fixed_point_factory: Callable[[float], FixedPoint]
) -> tuple[int, int]:
    dummy_value = fixed_point_factory(0)
    return dummy_value.total_bits, dummy_value.frac_bits


def calculate_address_width(num_items: int) -> int:
    return max(1, math.ceil(math.log2(num_items)))



import unittest

import torch

from elasticai.creator.nn.hard_sigmoid import HardSigmoid
from elasticai.creator.nn.linear import FixedPointLinear
from elasticai.creator.nn.relu import ReLU
from elasticai.creator.vhdl.number_representations import FixedPoint, FixedPointFactory
from elasticai.creator.vhdl.translator.pytorch import translator


class FixedPointModel(torch.nn.Module):
    def __init__(self, fixed_point_factory: FixedPointFactory) -> None:
        super().__init__()

        self.linear1 = FixedPointLinear(
            in_features=3,
            out_features=4,
            bias=True,
            fixed_point_factory=fixed_point_factory,
        )
        self.linear2 = FixedPointLinear(
            in_features=4,
            out_features=1,
            bias=True,
            fixed_point_factory=fixed_point_factory,
        )
        self.hard_sigmoid = HardSigmoid()
        self.relu1 = ReLU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.linear1(x)
        x = self.relu1(x)
        x = self.linear2(x)
        x = self.hard_sigmoid(x)
        return x


class TranslateLinearModelTest(unittest.TestCase):
    def setUp(self) -> None:
        torch.manual_seed(22)
        torch.cuda.manual_seed_all(22)

        fixed_point_factory = FixedPoint.get_factory(total_bits=8, frac_bits=4)
        model = FixedPointModel(fixed_point_factory)

        translation_args = {
            "0": {"work_library_name": "work"},
            "1": {"work_library_name": "work"},
            "2": {"fixed_point_factory": fixed_point_factory},
            "3": {"fixed_point_factory": fixed_point_factory},
        }
        code_repr = translator.translate_model(model, translation_args=translation_args)
        self.code_modules = list(code_repr)

    def test_code_structure_matches_expected(self) -> None:
        actual_code_structure = [
            (code_module.name, {file.name for file in code_module.files})
            for code_module in self.code_modules
        ]

        expected_code_structure = [
            (
                "0_FixedPointLinear",
                {
                    "fp_linear_1d_0.vhd",
                    "w_rom_fp_linear_1d_0.vhd",
                    "b_rom_fp_linear_1d_0.vhd",
                },
            ),
            (
                "1_FixedPointLinear",
                {
                    "fp_linear_1d_1.vhd",
                    "w_rom_fp_linear_1d_1.vhd",
                    "b_rom_fp_linear_1d_1.vhd",
                },
            ),
            ("2_HardSigmoid", {"fp_hard_sigmoid.vhd"}),
            ("3_ReLU", {"fp_relu_3.vhd"}),
            ("network_component", {"network.vhd"}),
        ]

        self.assertEqual(len(actual_code_structure), len(expected_code_structure))

        for expected_name, expected_files in expected_code_structure:
            with self.subTest(f"module {expected_name} in actual"):
                self.assertIn((expected_name, expected_files), actual_code_structure)



from unittest import TestCase

from elasticai.creator.vhdl.code_files.dual_port_2_clock_ram_component import (
    DualPort2ClockRamVHDLFile,
)
from elasticai.creator.vhdl.code_files.fp_hard_sigmoid_file import FPHardSigmoidFile
from elasticai.creator.vhdl.code_files.fp_hard_tanh_component import FPHardTanhComponent
from elasticai.creator.vhdl.code_files.lstm_component import LSTMFile
from elasticai.creator.vhdl.code_files.rom_component import RomFile
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.translator.abstract.layers import LSTMModule


class LSTMModuleTest(TestCase):
    def setUp(self) -> None:
        self.lstm = LSTMModule(
            weights_ih=[[[1, 2], [3, 4], [5, 6], [7, 8]]],
            weights_hh=[[[1], [2], [3], [4]]],
            biases_ih=[[1, 2, 3, 4]],
            biases_hh=[[5, 6, 7, 8]],
            layer_id="0",
            fixed_point_factory=FixedPoint.get_factory(total_bits=8, frac_bits=2),
            work_library_name="work",
        )

    def test_correct_number_of_components(self) -> None:
        vhdl_components = list(self.lstm.files)
        self.assertEqual(len(vhdl_components), 12)

    def test_contains_all_needed_components(self) -> None:
        vhdl_components = self.lstm.files

        target_components = [
            (RomFile, "wi_rom.vhd"),
            (RomFile, "wf_rom.vhd"),
            (RomFile, "wg_rom.vhd"),
            (RomFile, "wo_rom.vhd"),
            (RomFile, "bi_rom.vhd"),
            (RomFile, "bf_rom.vhd"),
            (RomFile, "bg_rom.vhd"),
            (RomFile, "bo_rom.vhd"),
            (FPHardSigmoidFile, "fp_hard_sigmoid.vhd"),
            (FPHardTanhComponent, "fp_hard_tanh.vhd"),
            (LSTMFile, "lstm.vhd"),
            (DualPort2ClockRamVHDLFile, "dual_port_2_clock_ram.vhd"),
        ]
        actual_components = [(type(x), x.name) for x in vhdl_components]

        self.assertEqual(target_components, actual_components)



import typing
import unittest

from elasticai.creator.mlframework import Module
from elasticai.creator.vhdl.hw_equivalent_layers.layers import (
    FixedPointLinear,
    RootModule,
)
from elasticai.creator.vhdl.model_tracing import HWEquivalentNode, HWEquivalentTracer
from elasticai.creator.vhdl.number_representations import ClippedFixedPoint


class FPLinearModel(RootModule):
    def __init__(self):
        super().__init__()
        data_width = 16
        fp_factory = ClippedFixedPoint.get_factory(total_bits=data_width, frac_bits=8)
        self.fp_linear = FixedPointLinear(
            in_features=1,
            out_features=1,
            fixed_point_factory=fp_factory,
            data_width=data_width,
        )

    def forward(self, x):
        return self.fp_linear(x)


class FPLinearModelWithGetItemFunctionNode(FPLinearModel):
    def forward(self, x):
        return self.fp_linear(x[:2])


class TestForTracingHWEquivalentModelsToGenerateVHDL(unittest.TestCase):
    def test_generated_graph_nodes_provide_corresponding_modules(self):
        model = FPLinearModel()
        tracer = HWEquivalentTracer()
        graph = tracer.trace(model)
        layer = None
        for node in graph.nodes:
            if isinstance(node, HWEquivalentNode):
                layer = node.hw_equivalent_layer
        self.assertTrue(
            layer is model.fp_linear,
            "expected: {}, actual: {}".format(layer, model.fp_linear),
        )

    def test_graph_for_single_fp_linear_layer_model_contains_no_call_function_nodes(
        self,
    ):
        model = FPLinearModel()
        tracer = HWEquivalentTracer()
        graph = tracer.trace(typing.cast(Module, model))
        self.assertFalse(
            any(n.op == "call_function" for n in graph.nodes),
            "; ".join((f"({n.name}, {n.op})" for n in graph.nodes)),
        )

    def test_graph_provides_exactly_one_module_node(self):
        model = FPLinearModelWithGetItemFunctionNode()
        graph = HWEquivalentTracer().trace(typing.cast(Module, model))
        nodes = tuple(graph.hw_equivalent_nodes)
        self.assertEqual(model.fp_linear, nodes[0].hw_equivalent_layer)
        self.assertEqual(1, len(nodes))



from elasticai.creator.nn.hard_sigmoid import HardSigmoid
from elasticai.creator.vhdl.number_representations import FixedPointFactory
from elasticai.creator.vhdl.translator.abstract.layers.fp_hard_sigmoid_module import (
    FPHardSigmoidModule,
)


def build_fp_hard_sigmoid(
    layer: HardSigmoid, layer_id: str, fixed_point_factory: FixedPointFactory
) -> FPHardSigmoidModule:
    return FPHardSigmoidModule(
        layer_id=layer_id, fixed_point_factory=fixed_point_factory
    )



from elasticai.creator.vhdl.number_representations import FixedPoint, FixedPointFactory
from elasticai.creator.vhdl.vhdl_files import VHDLFile


class FPHardTanhComponent(VHDLFile):
    def __init__(
        self,
        min_val: FixedPoint,
        max_val: FixedPoint,
        fixed_point_factory: FixedPointFactory,
        layer_id: str,
    ):
        d = dict(
            data_width=fixed_point_factory.total_bits,
            frac_width=fixed_point_factory.frac_bits,
            min_val=min_val.to_signed_int(),
            max_val=max_val.to_signed_int(),
            layer_name=layer_id,
        )
        stringified_d = dict(((k, str(v)) for k, v in d.items()))
        super().__init__(name="fp_hard_tanh", **stringified_d)



from elasticai.creator.vhdl.vhdl_files import VHDLFile


class DualPort2ClockRamVHDLFile(VHDLFile):
    def __init__(self, layer_id: str):
        super().__init__(name="dual_port_2_clock_ram", layer_name=layer_id)



import unittest
from io import StringIO
from typing import Iterable

from elasticai.creator.tests.integration.vhdl.vhd_file_reader import (
    VHDLFileReaderWithoutComments,
)
from elasticai.creator.vhdl.code import Code, CodeModule


class CodeTestCase(unittest.TestCase):
    @staticmethod
    def unified_vhdl_from_module(module: CodeModule):
        vhdl_file = next(iter(module.files))
        code = "\n".join(vhdl_file.code())
        io = StringIO(code)
        codes = VHDLFileReaderWithoutComments(io).as_list()
        return codes

    @staticmethod
    def code_section_from_string(s: str) -> list[Code]:
        return [CodeTestCase.code_from_string(s)]

    @staticmethod
    def code_from_string(s: str) -> Code:
        return VHDLFileReaderWithoutComments(StringIO(s))

    def check_contains_all_expected_lines(self, expected: Code, actual: Code):
        reusable_code = list(actual)
        for line in expected:
            with self.subTest(line):
                self.assertTrue(
                    line in reusable_code,
                    f"expected to find: {line}\nbut found: {reusable_code}",
                )

    def check_lines_are_equal_ignoring_order(
        self, expected: Iterable[Code], actual: Iterable[Code]
    ):
        def unpack(codes: Iterable[Code]) -> list[list[str]]:
            return list(map(lambda x: list(x), codes))

        def sort(codes: Iterable[Code]) -> Iterable[Code]:
            return sorted(map(lambda x: sorted(x), codes))

        actual = unpack(sort(actual))
        expected = unpack(sort(expected))
        self.assertEqual(expected, actual)



import dataclasses
from abc import abstractmethod
from collections.abc import Collection
from typing import Callable, Iterable, Optional, Protocol, Union, runtime_checkable

Code = Iterable[str]


@runtime_checkable
class CodeGenerator(Protocol):
    @abstractmethod
    def code(self) -> Code:
        ...


CodeGeneratorCompatible = Code | CodeGenerator | str | Callable[[], Code]


class Translatable(Protocol):
    @abstractmethod
    def translate(self) -> "CodeModule":
        ...


class CodeFile(Protocol):
    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @abstractmethod
    def code(self) -> Code:
        ...


class TemplateCodeFile(CodeFile, Protocol):
    @property
    @abstractmethod
    def multi_line_parameters(self) -> dict[str, Iterable[str]]:
        ...

    @property
    @abstractmethod
    def single_line_parameters(self) -> dict[str, str]:
        ...

    @property
    @abstractmethod
    def parameters(self) -> dict[str, Union[str, Iterable[str]]]:
        ...


class CodeModule(Protocol):
    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @property
    @abstractmethod
    def files(self) -> Collection[CodeFile]:
        ...

    @property
    @abstractmethod
    def submodules(self) -> Collection["CodeModule"]:
        ...


@dataclasses.dataclass
class CodeModuleBase(CodeModule):
    @property
    def submodules(self) -> Collection["CodeModule"]:
        return self._submodules

    @property
    def files(self) -> Collection[CodeFile]:
        return self._files

    @property
    def name(self) -> str:
        return self._name

    def __init__(
        self,
        name: str,
        files: Collection[CodeFile],
        submodules: Collection["CodeModule"] = tuple(),
    ):
        self._name = name
        self._files = files
        self._submodules = submodules


class CodeFileBase(CodeFile):
    @property
    def name(self) -> str:
        return self._name

    def code(self) -> Code:
        return self._code

    def __repr__(self) -> str:
        return f"CodeBaseFile(name={self._name}, code={self._code})"

    def __init__(self, name: str, code: Code):
        self._name = name
        self._code = code



"""
The module contains classes and functions for generating vhdl code.
We provide code generators for the subset of vhdl that we need to implement
our neural network_component.py accelerators and test benches. We stick closely to the vhdl
formal grammar with our class names.

The core of this module is the `CodeGenerator`. Code generators are callables that return `Code`.
`Code` is an Iterable of strings. Depending on complexity, we define syntactic code_files of the vhdl
grammar as `CodeGenerator`s. The class can then be used to set up and configure a function that yields lines
of code as strings.
"""
from collections.abc import Iterable
from enum import Enum
from itertools import filterfalse
from typing import Callable, Iterator, Literal, Optional, Sequence, overload

from elasticai.creator.vhdl.code import Code, CodeGenerator, CodeGeneratorCompatible

Identifier = str


class Keywords(Enum):
    IS = "is"
    END = "end"
    ENTITY = "entity"
    COMPONENT = "component"
    PORT = "port"
    GENERIC = "generic"
    IN = "in"
    OUT = "out"
    INOUT = "inout"
    BUFFER = "buffer"
    LINKAGE = "linkage"
    INTEGER = "integer"
    STD_LOGIC = "std_logic"
    SIGNAL = "signal"
    ARCHITECTURE = "architecture"
    OF = "of"
    SIGNED = "signed"
    BEGIN = "begin"
    PROCESS = "process"
    STD_LOGIC_VECTOR = "std_logic_vector"


class DataType(Enum):
    INTEGER = Keywords.INTEGER.value
    STD_LOGIC = Keywords.STD_LOGIC.value
    SIGNED = Keywords.SIGNED.value
    STD_LOGIC_VECTOR = Keywords.STD_LOGIC_VECTOR.value


class Mode(Enum):
    IN = Keywords.IN.value
    OUT = Keywords.OUT.value
    INOUT = Keywords.INOUT.value
    BUFFER = Keywords.BUFFER.value


class _DesignUnitForEntityAndComponent:
    def __init__(
        self, identifier: str, design_type: Literal[Keywords.ENTITY, Keywords.COMPONENT]
    ):
        self.identifier = identifier
        self._generic_list = InterfaceList()
        self._port_list = InterfaceList()
        self.type = design_type

    @property
    def generic_list(self):
        return self._generic_list

    @generic_list.setter
    def generic_list(self, value):
        self._generic_list = InterfaceList(value)

    @property
    def port_list(self):
        return self._port_list

    @port_list.setter
    def port_list(self, value):
        self._port_list = InterfaceList(value)

    def _header(self) -> Iterator[str]:
        if len(self.generic_list) > 0:
            yield from _clause(Keywords.GENERIC, self._generic_list.code())
        if len(self.port_list) > 0:
            yield from _clause(Keywords.PORT, self._port_list.code())

    def __iter__(self) -> Iterator[str]:
        yield f"{self.type.value} {self.identifier} {Keywords.IS.value}"
        yield from _filter_empty_lines(self._header())
        yield f"{Keywords.END.value} {self.type.value} {self.identifier};"

    def code(self) -> Code:
        return self


class Entity(_DesignUnitForEntityAndComponent):
    def __init__(self, identifier: str):
        super().__init__(identifier, Keywords.ENTITY)


class ComponentDeclaration(_DesignUnitForEntityAndComponent):
    def __init__(self, identifier: str):
        super().__init__(identifier, Keywords.COMPONENT)


class Architecture:
    def __init__(self, design_unit: str):
        self.design_unit = design_unit
        self.architecture_declaration_list = InterfaceList()
        self.architecture_component_list = InterfaceList()
        self.architecture_assignment_list = InterfaceList()
        self.architecture_process_list = InterfaceList()
        self.architecture_port_map_list = InterfaceList()
        self.architecture_assignment_at_end_of_declaration_list = InterfaceList()
        self.architecture_statement_part: Optional[CodeGenerator] = None

    def _generator(self) -> Iterator[str]:
        yield (
            f"{Keywords.ARCHITECTURE.value} rtl"
            f" {Keywords.OF.value} {self.design_unit} {Keywords.IS.value}"
        )
        if len(self.architecture_declaration_list) > 0:
            yield from _filter_empty_lines(
                _add_semicolons(
                    self.architecture_declaration_list.code(), semicolon_last=True
                )
            )
        if len(self.architecture_component_list) > 0:
            yield from _filter_empty_lines(self.architecture_component_list.code())
        yield f"{Keywords.BEGIN.value}"
        if len(self.architecture_assignment_list) > 0:
            yield from _filter_empty_lines(
                _add_semicolons(
                    self.architecture_assignment_list.code(), semicolon_last=True
                )
            )
        if len(self.architecture_process_list) > 0:
            yield from _filter_empty_lines(self.architecture_process_list.code())
        if len(self.architecture_port_map_list) > 0:
            yield from _filter_empty_lines(self.architecture_port_map_list.code())
        if len(self.architecture_assignment_at_end_of_declaration_list) > 0:
            yield from _filter_empty_lines(
                _add_semicolons(
                    self.architecture_assignment_at_end_of_declaration_list.code(),
                    semicolon_last=True,
                )
            )

        if self.architecture_statement_part:
            yield from _filter_empty_lines(self.architecture_statement_part.code())
        yield f"{Keywords.END.value} {Keywords.ARCHITECTURE.value} rtl;"

    def code(self) -> Code:
        return self._generator()


class Process(Code, CodeGenerator):
    def __init__(
        self,
        identifier: str,
        input_name: Optional[str] = None,
        lookup_table: Optional[Code] = None,
    ):
        self.identifier = identifier
        self.process_declaration_list: list[str] = []
        self.process_statements_list: list[str] = []
        self.lookup_table = lookup_table
        self.input = input_name

    def _header(self) -> Iterator[str]:
        if len(self.process_declaration_list) > 0:
            yield from _append_semicolons_to_lines(self.process_declaration_list)

    def _body(self) -> Iterator[str]:
        if len(self.process_statements_list) > 0:
            yield from _append_semicolons_to_lines(self.process_statements_list)
        if self.lookup_table:
            yield from self.lookup_table

    def __iter__(self) -> Iterator[str]:
        if self.input:
            yield (
                f"{self.identifier}_{Keywords.PROCESS.value}:"
                f" {Keywords.PROCESS.value}({self.input})"
            )
        else:
            yield (
                f"{self.identifier}_{Keywords.PROCESS.value}: {Keywords.PROCESS.value}"
            )
        yield from _filter_empty_lines(self._header())
        yield f"{Keywords.BEGIN.value}"
        yield from _filter_empty_lines(self._body())
        yield (
            f"{Keywords.END.value} {Keywords.PROCESS.value} {self.identifier}_{Keywords.PROCESS.value};"
        )

    def code(self):
        return self


class ContextClause(Code, CodeGenerator):
    def __init__(self, library_clause, use_clause):
        self._use_clause = use_clause
        self._library_clause = library_clause

    def __iter__(self) -> Iterator[str]:
        yield from self._library_clause
        yield from self._use_clause

    def code(self) -> Code:
        return self


class UseClause(Code, CodeGenerator):
    def __init__(self, selected_names: list[str]):
        self._selected_names = selected_names

    def __iter__(self):
        def prefix_use(line: str):
            return f"use {line}"

        yield from _append_semicolons_to_lines(map(prefix_use, self._selected_names))

    def code(self) -> Code:
        return self


class LibraryClause(Code, CodeGenerator):
    def __init__(self, logical_name_list: list[str]):
        self._logical_name_list = logical_name_list

    def __iter__(self):
        yield from _append_semicolons_to_lines(
            ["library {}".format(", ".join(self._logical_name_list))]
        )

    def code(self) -> Code:
        return self


class InterfaceConstrained(Code, CodeGenerator):
    def __init__(
        self,
        identifier: str,
        identifier_type: DataType,
        width: Optional[str | int],
        mode: Optional[Mode],
        value: Optional[str | int],
        declaration_type: Optional[str],
    ):
        self._identifier = identifier
        self._range = f"({width})" if width else ""
        self._identifier_type = identifier_type
        self._mode = f" {mode.value} " if mode else " "
        self._value = f" := {value}" if value else ""
        self._declaration_type = f"{declaration_type} " if declaration_type else ""

    def __iter__(self) -> Iterator[str]:
        yield from (
            (
                f"{self._declaration_type}{self._identifier} :"
                f"{self._mode}{self._identifier_type.value}{self._range}{self._value}"
            ),
        )

    def code(self) -> Code:
        return self


class InterfaceVariable(InterfaceConstrained):
    def __init__(
        self,
        identifier: str,
        identifier_type: DataType,
        width: Optional[str | int] = None,
        mode: Optional[Mode] = None,
        value: Optional[str | int] = None,
    ):
        super().__init__(
            identifier, identifier_type, width, mode, value, declaration_type=None
        )


def _wrap_iterator_function_into_iterable(create_iterator: Callable[[], Iterator[str]]):
    class Wrapper(Iterable[str]):
        def __iter__(self) -> Iterator[str]:
            yield from create_iterator()

    wrapped = Wrapper()
    return wrapped


class CodeGeneratorConcatenation(Sequence[CodeGenerator], CodeGenerator):
    @overload
    def __getitem__(self, index: int) -> CodeGenerator:
        ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[CodeGenerator]:
        ...

    def __getitem__(self, index):
        if isinstance(index, slice):
            return CodeGeneratorConcatenation(self.interface_generators[index])
        else:
            return self.interface_generators[index]

    def __len__(self) -> int:
        return len(self.interface_generators)

    def __init__(self, *interfaces: CodeGeneratorCompatible):
        self.interface_generators: list[CodeGenerator] = [
            _unify_code_generators(interface) for interface in interfaces
        ]

    def append(self, interface: CodeGeneratorCompatible) -> None:
        self.interface_generators.append(_unify_code_generators(interface))

    def _code_iterator(self) -> Iterator[str]:
        for generator in self.interface_generators:
            for line in generator.code():
                yield line

    def code(self) -> Code:
        return _wrap_iterator_function_into_iterable(self._code_iterator)


class InterfaceSignal(InterfaceConstrained):
    def __init__(
        self,
        identifier: str,
        identifier_type: DataType,
        width: Optional[str | int] = None,
        mode: Optional[Mode] = None,
        value: Optional[str | int] = None,
    ) -> None:
        super().__init__(
            identifier, identifier_type, width, mode, value, declaration_type="signal"
        )


class PortMap:
    def __init__(self, map_name, component_name):
        self.map_name = map_name
        self.component_name = component_name
        self._signal_list = InterfaceList()
        self._generic_map_list = InterfaceList()

    @property
    def signal_list(self):
        return self._signal_list

    @signal_list.setter
    def signal_list(self, value):
        self._signal_list = InterfaceList(value)

    @property
    def generic_map_list(self):
        return self._generic_map_list

    @generic_map_list.setter
    def generic_map_list(self, value):
        self._generic_map_list = InterfaceList(value)

    def code(self) -> Code:
        def generator():
            yield f"{self.map_name}: {self.component_name}"
            if len(self._generic_map_list) > 0:
                yield f"generic map ("
                yield from _filter_empty_lines(
                    _add_comma(self._generic_map_list.code())
                )
                yield ")"
            yield f"port map ("
            yield from _filter_empty_lines(_add_comma(self._signal_list.code()))
            yield ");"

        return generator()


class InterfaceList(CodeGeneratorConcatenation):
    pass


ClauseType = Literal[Keywords.GENERIC, Keywords.COMPONENT, Keywords.PORT]


def indent(line: str) -> str:
    return "".join(["\t", line])


def _add_semicolons(lines: Code, semicolon_last: bool = False) -> Iterator[str]:
    temp = tuple(lines)
    yield from (f"{line};" for line in temp[:-1])
    yield f"{temp[-1]};" if semicolon_last else f"{temp[-1]}"


def _add_is(lines: Code) -> Iterator[str]:
    yield from (f"{line} is" for line in lines)


def _add_comma(lines: Code, comma_last: bool = False) -> Iterator[str]:
    temp = tuple(lines)
    yield from (f"{line}," for line in temp[:-1])
    yield f"{temp[-1]}," if comma_last else f"{temp[-1]}"


def _append_semicolons_to_lines(lines: Code) -> Iterator[str]:
    temp = tuple(lines)
    yield from (f"{line};" for line in temp)


def _clause(clause_type: ClauseType, interfaces: Code) -> Iterator[str]:
    yield f"{clause_type.value} ("
    yield from _filter_empty_lines(_add_semicolons(interfaces))
    yield ");"


def _filter_empty_lines(lines: Code) -> Iterator[str]:
    return filterfalse(_line_is_empty, lines)


def _line_is_empty(line: str) -> bool:
    return len(line) == 0


def _join_lines(lines) -> str:
    return "\n".join(lines)


def _empty_code_generator() -> Code:
    return []


def _wrap_in_is_end_block(
    block_type: Keywords, block_identifier: Identifier, lines: Code
) -> Iterator[str]:
    yield f"{block_type.value} {block_identifier} {Keywords.IS.value}"
    yield from _filter_empty_lines(lines)
    yield f"{Keywords.END.value} {block_type.value} {block_identifier};"


def _wrap_string_into_code_generator(string: str) -> CodeGenerator:
    class Wrapper(CodeGenerator):
        def code(self) -> Code:
            return [string]

    return Wrapper()


def _wrap_code_into_code_generator(code: Code) -> CodeGenerator:
    class Wrapped(CodeGenerator):
        def code(self) -> Code:
            return code

    wrapped = Wrapped()
    return wrapped


def _wrap_callable_into_code_generator(fn: Callable[[], Code]) -> CodeGenerator:
    class Wrapper(CodeGenerator):
        def code(self) -> Code:
            return fn()

    wrapped = Wrapper()
    return wrapped


def _unify_code_generators(generator: CodeGeneratorCompatible) -> CodeGenerator:
    if isinstance(generator, str):
        return _wrap_string_into_code_generator(generator)
    elif isinstance(generator, Iterable):
        return _wrap_code_into_code_generator(generator)
    elif isinstance(generator, CodeGenerator):
        return generator
    elif callable(generator):
        return _wrap_callable_into_code_generator(generator)
    else:
        raise ValueError


class Procedure:
    def __init__(self, identifier: str):
        self.identifier = identifier
        self._declaration_list = InterfaceList()
        self._declaration_list_with_is = InterfaceList()
        self._statement_list = InterfaceList()

    @property
    def declaration_list(self):
        return self._declaration_list

    @declaration_list.setter
    def declaration_list(self, value):
        self._declaration_list = InterfaceList(value)

    @property
    def declaration_list_with_is(self):
        return self._declaration_list_with_is

    @declaration_list_with_is.setter
    def declaration_list_with_is(self, value):
        self._declaration_list_with_is = InterfaceList(value)

    @property
    def statement_list(self):
        return self._statement_list

    @statement_list.setter
    def statement_list(self, value):
        self._statement_list = InterfaceList(value)

    def _generator(self):
        yield f"procedure {self.identifier} ("
        if len(self._declaration_list) > 0:
            yield from _filter_empty_lines(
                _add_semicolons(self._declaration_list.code(), semicolon_last=True)
            )
        if len(self._declaration_list_with_is) > 0:
            yield from _filter_empty_lines(
                _add_is(self._declaration_list_with_is.code())
            )
        yield f"{Keywords.BEGIN.value}"
        if len(self._statement_list) > 0:
            yield from _filter_empty_lines(
                _add_semicolons(self._statement_list.code(), semicolon_last=True)
            )
        yield f"{Keywords.END.value} {self.identifier};"

    def code(self) -> Code:
        return self._generator()


def hex_representation(hex_value: str) -> str:
    return f'x"{hex_value}"'


def bin_representation(bin_value: str) -> str:
    return f'"{bin_value}"'



from collections.abc import Callable
from functools import partial
from typing import Optional, cast

import torch

from elasticai.creator.nn.arithmetics import FixedPointArithmetics
from elasticai.creator.nn.hard_sigmoid import HardSigmoid
from elasticai.creator.nn.hard_tanh import HardTanh
from elasticai.creator.nn.lstm_cell import LSTMCell
from elasticai.creator.vhdl.number_representations import FixedPointFactory


class LSTM(torch.nn.Module):
    def __init__(
        self,
        input_size: int,
        hidden_size,
        bias: bool,
        batch_first: bool,
        lstm_cell_factory: Callable[..., torch.nn.Module],
    ) -> None:
        super().__init__()
        self.cell = lstm_cell_factory(
            input_size=input_size, hidden_size=hidden_size, bias=bias
        )
        self.batch_first = batch_first

    def forward(
        self,
        x: torch.Tensor,
        state: Optional[tuple[torch.Tensor, torch.Tensor]] = None,
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, torch.Tensor]]:
        batched = x.dim() == 3

        if batched and self.batch_first:
            x = torch.stack(torch.unbind(x), dim=1)

        if state is not None:
            state = state[0].squeeze(0), state[1].squeeze(0)

        inputs = torch.unbind(x, dim=0)

        outputs = []
        for i in range(len(inputs)):
            hidden_state, cell_state = self.cell(inputs[i], state)
            state = (hidden_state, cell_state)
            outputs.append(hidden_state)

        if state is None:
            raise RuntimeError("Number of samples must be larger than 0.")

        result = torch.stack(outputs, dim=1 if batched and self.batch_first else 0)
        hidden_state, cell_state = state[0].unsqueeze(0), state[1].unsqueeze(0)

        return result, (hidden_state, cell_state)


class FixedPointLSTMWithHardActivations(LSTM):
    def __init__(
        self,
        fixed_point_factory: FixedPointFactory,
        input_size: int,
        hidden_size: int,
        batch_first: bool,
        bias: bool = True,
    ) -> None:
        super().__init__(
            input_size=input_size,
            hidden_size=hidden_size,
            bias=bias,
            batch_first=batch_first,
            lstm_cell_factory=partial(
                LSTMCell,
                arithmetics=FixedPointArithmetics(
                    fixed_point_factory=fixed_point_factory
                ),
                sigmoid_factory=HardSigmoid,
                tanh_factory=HardTanh,
            ),
        )

    @property
    def fixed_point_factory(self) -> FixedPointFactory:
        return cast(FixedPointArithmetics, self.cell.ops).fixed_point_factory



from dataclasses import dataclass
from typing import Collection

from elasticai.creator.vhdl.code import CodeFile, CodeModule
from elasticai.creator.vhdl.code_files.fp_relu_component import FPReLUComponent
from elasticai.creator.vhdl.number_representations import FixedPointFactory


@dataclass
class FPReLUModule(CodeModule):
    layer_id: str
    fixed_point_factory: FixedPointFactory

    @property
    def submodules(self) -> Collection["CodeModule"]:
        raise NotImplementedError()

    @property
    def name(self) -> str:
        return self.layer_id

    @property
    def files(self) -> Collection[CodeFile]:
        yield FPReLUComponent(
            layer_id=self.layer_id,
            fixed_point_factory=self.fixed_point_factory,
        )



from math import ceil, floor

import torch
from matplotlib import pyplot as plt
from torch.utils.data import DataLoader, Dataset, random_split

from elasticai.creator.nn.linear import FixedPointLinear
from elasticai.creator.nn.lstm import FixedPointLSTMWithHardActivations
from elasticai.creator.vhdl.number_representations import FixedPoint


def _x_values(value_range: tuple[float, float], sampling_rate: int) -> torch.Tensor:
    start, end = value_range
    return torch.linspace(
        start=start, end=end, steps=int((end - start) * sampling_rate)
    )


class SineDataset(Dataset):
    def __init__(
        self, sine_range: tuple[float, float], sampling_rate: int, window_size: int
    ) -> None:
        self.sine_range = sine_range
        self.sampling_rate = sampling_rate
        self.window_size = window_size

        x = _x_values(self.sine_range, self.sampling_rate)
        self._sinus = torch.sin(x)
        self._samples, self._labels = self._extract_sample_label_pairs()

    def _extract_sample_label_pairs(self) -> tuple[torch.Tensor, torch.Tensor]:
        samples, labels = [], []
        for i in range(self.window_size, len(self._sinus)):
            samples.append(self._sinus[i - self.window_size : i].view(-1, 1).tolist())
            labels.append(self._sinus[i].item())
        return (
            torch.tensor(samples, dtype=self._sinus.dtype),
            torch.tensor(labels, dtype=self._sinus.dtype),
        )

    @property
    def sinus(self):
        return self._sinus

    def __len__(self) -> int:
        return len(self._labels)

    def __getitem__(self, idx) -> tuple[torch.Tensor, torch.Tensor]:
        return self._samples[idx], self._labels[idx]


def train(
    model: torch.nn.Module,
    train_data: Dataset,
    test_data: Dataset,
    batch_size: int,
    learning_rate: float,
    epochs: int,
) -> None:
    dl_train = DataLoader(train_data, batch_size=batch_size, shuffle=True)
    dl_test = DataLoader(test_data, batch_size=batch_size, shuffle=True)

    loss_fn = torch.nn.MSELoss()
    optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)

    for epoch in range(epochs):
        losses = []
        for samples, labels in dl_train:
            predictions = model(samples)
            loss = loss_fn(predictions, labels.view(-1, 1))
            loss.backward()
            optimizer.step()
            optimizer.zero_grad()
            losses.append(loss.item())
        train_loss = sum(losses) / len(losses)

        with torch.no_grad():
            val_losses = map(lambda x: loss_fn(model(x[0]), x[1].view(-1, 1)), dl_test)
            val_loss = sum(val_losses) / len(dl_test)

        print(f"[epoch: {epoch + 1}/{epochs}] loss: {train_loss}; val_loss: {val_loss}")

    model.eval()


def plot_predicted_sine(
    model: torch.nn.Module,
    prediction_range: tuple[float, float],
    sampling_rate: int,
    window_size: int,
) -> None:
    pred_start, pred_end = prediction_range
    x_prelude = _x_values(
        (pred_start - window_size / sampling_rate, pred_start), sampling_rate
    )
    x_to_predict = _x_values(prediction_range, sampling_rate)
    y_prelude = torch.sin(x_prelude)
    y_target = torch.sin(x_to_predict)

    predictions: list[float] = y_prelude.tolist()
    for i in range(int((pred_end - pred_start) * sampling_rate)):
        inputs = torch.tensor(predictions[i:]).view(-1, 1)
        prediction = model(inputs)
        predictions.append(prediction[0].item())
    y_pred = torch.tensor(predictions[window_size:])

    plt.plot(x_prelude, y_prelude, "b-", label="prelude")
    plt.plot(x_to_predict, y_target, "g-", label="target")
    plt.plot(x_to_predict, y_pred, "r--", label="prediction")
    plt.legend()
    plt.show()


class FixedPointSineModel(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self._hidden_size = 64
        self._fp_factory = FixedPoint.get_factory(total_bits=16, frac_bits=8)
        self.lstm = FixedPointLSTMWithHardActivations(
            input_size=1,
            hidden_size=self._hidden_size,
            bias=True,
            batch_first=True,
            fixed_point_factory=self._fp_factory,
        )
        self.linear = FixedPointLinear(
            in_features=self._hidden_size,
            out_features=1,
            bias=True,
            fixed_point_factory=self._fp_factory,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        _, (final_hidden_state, _) = self.lstm(x)
        final_hidden_state = final_hidden_state.squeeze(0)
        return self.linear(final_hidden_state)


def main() -> None:
    dataset = SineDataset(
        sine_range=(0, 10 * torch.pi), sampling_rate=16, window_size=32
    )
    ds_train, ds_test = random_split(
        dataset, [floor(len(dataset) * 0.8), ceil(len(dataset) * 0.2)]
    )
    model = FixedPointSineModel()

    train(
        model=model,
        train_data=ds_train,
        test_data=ds_test,
        batch_size=32,
        learning_rate=1e-3,
        epochs=100,
    )

    plot_predicted_sine(
        model=model,
        prediction_range=(0, 2 * torch.pi),
        sampling_rate=16,
        window_size=32,
    )


if __name__ == "__main__":
    main()



import torch

from elasticai.creator.nn.autograd_functions.fixed_point_quantization import (
    FixedPointDequantFunction,
    FixedPointQuantFunction,
)
from elasticai.creator.tests.unit.tensor_test_case import TensorTestCase
from elasticai.creator.vhdl.number_representations import FixedPoint


class FixedPointQuantFunctionTest(TensorTestCase):
    def setUp(self) -> None:
        self.fp_factory = FixedPoint.get_factory(total_bits=4, frac_bits=2)
        self.quant = lambda x: FixedPointQuantFunction.apply(x, self.fp_factory)

    def test_quantize_upper_bound(self) -> None:
        x = torch.tensor([1.75])
        actual = self.quant(x)
        target = [7.0]
        self.assertTensorEqual(target, actual)

    def test_quantize_lower_bound(self) -> None:
        x = torch.tensor([-2.0])
        actual = self.quant(x)
        target = [-8.0]
        self.assertTensorEqual(target, actual)

    def test_quantize_out_of_bounds(self) -> None:
        x = torch.tensor([2.0])
        with self.assertRaises(ValueError):
            _ = self.quant(x)

    def test_quantize_typical_values(self) -> None:
        x = torch.tensor([-1.3, 0.1, 1.6])
        actual = self.quant(x)
        target = [-5, 0, 6]
        self.assertTensorEqual(target, actual)


class FixedPointDequantFunctionTest(TensorTestCase):
    def setUp(self) -> None:
        self.fp_factory = FixedPoint.get_factory(total_bits=4, frac_bits=2)
        self.dequant = lambda x: FixedPointDequantFunction.apply(x, self.fp_factory)

    def test_dequantize_upper_bound(self) -> None:
        x = torch.tensor([7.0])
        actual = self.dequant(x)
        target = [1.75]
        self.assertTensorEqual(target, actual)

    def test_dequantize_lower_bound(self) -> None:
        x = torch.tensor([-8.0])
        actual = self.dequant(x)
        target = [-2.0]
        self.assertTensorEqual(target, actual)

    def test_dequantize_out_of_bounds(self) -> None:
        x = torch.tensor([9])
        with self.assertRaises(ValueError):
            _ = self.dequant(x)

    def test_dequantize_typical_values(self) -> None:
        x = torch.tensor([-5, 0, 6])
        actual = self.dequant(x)
        target = [-1.25, 0, 1.5]
        self.assertTensorEqual(target, actual)



from dataclasses import dataclass

from elasticai.creator.resource_utils import read_text
from elasticai.creator.vhdl.code import Code, CodeFile
from elasticai.creator.vhdl.number_representations import FixedPointFactory
from elasticai.creator.vhdl.vhdl_files import expand_template


@dataclass
class FPReLUComponent(CodeFile):
    layer_id: str
    fixed_point_factory: FixedPointFactory

    def __post_init__(self) -> None:
        self.data_width = self.fixed_point_factory.total_bits
        self.frac_width = self.fixed_point_factory.frac_bits

    @property
    def name(self) -> str:
        return f"fp_relu_{self.layer_id}.vhd"

    def code(self) -> Code:
        template = read_text("elasticai.creator.vhdl.templates", "fp_relu.tpl.vhd")

        code = expand_template(
            template,
            layer_name=self.layer_id,
            data_width=str(self.data_width),
            clock_option="false",
        )

        return code



from typing import Protocol

from elasticai.creator.mlframework import Module
from elasticai.creator.vhdl.code import Translatable
from elasticai.creator.vhdl.hw_equivalent_layers.hw_blocks import HWBlockInterface


class HWEquivalentLayer(HWBlockInterface, Translatable, Module, Protocol):
    ...



from typing import Union

from elasticai.creator.vhdl.code import Code


def extract_section(
    begin: Union[str, Code], end: Union[str, Code], lines: Code
) -> list[list[str]]:
    extract = False
    content: list[list[str]] = []
    current_section: list[str] = []
    begin = list(begin) if not isinstance(begin, str) else [begin]
    end = list(end) if not isinstance(end, str) else [end]
    look_back = len(begin)
    look_ahead = len(end)
    lines = list(lines)
    i = 0
    last_i = len(lines)
    while i < last_i:
        look_ahead_window = lines[i : i + look_ahead]
        look_back_window = lines[i : i + look_back]

        if not extract and look_back_window == begin:
            extract = True
            i = i + look_back - 1
        elif extract and look_ahead_window == end:
            extract = False
            content.append(current_section)
            current_section = []
        elif extract:
            current_section.append(lines[i])
        i += 1

    if extract:
        raise ValueError(f"reached end of code before end: {end}")
    return content



from typing import Any, Optional

import torch

from elasticai.creator.nn.arithmetics import FloatArithmetics
from elasticai.creator.nn.lstm import LSTM
from elasticai.creator.nn.lstm_cell import LSTMCell
from elasticai.creator.tests.unit.tensor_test_case import TensorTestCase


def create_lstm(
    input_size: int, hidden_size: int, bias: bool, batch_first: bool
) -> tuple[torch.nn.Module, torch.nn.Module]:
    lstm_args: dict[str, Any] = dict(
        input_size=input_size,
        hidden_size=hidden_size,
        bias=bias,
        batch_first=batch_first,
    )
    lstm = LSTM(**lstm_args, lstm_cell_factory=torch.nn.LSTMCell)
    reference_lstm = torch.nn.LSTM(**lstm_args)

    lstm.cell.weight_ih = reference_lstm.weight_ih_l0
    lstm.cell.weight_hh = reference_lstm.weight_hh_l0
    if bias:
        lstm.cell.bias_ih = reference_lstm.bias_ih_l0
        lstm.cell.bias_hh = reference_lstm.bias_hh_l0

    return lstm, reference_lstm


class OutputsZeroLSTMCell(LSTMCell):
    def __init__(self, input_size: int, hidden_size: int, bias: bool) -> None:
        super().__init__(
            input_size=input_size,
            hidden_size=hidden_size,
            bias=bias,
            arithmetics=FloatArithmetics(),
            sigmoid_factory=torch.nn.Sigmoid,
            tanh_factory=torch.nn.Tanh,
        )

    def forward(
        self, x: torch.Tensor, state: Optional[tuple[torch.Tensor, torch.Tensor]] = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        h, c = super().forward(x, state)
        return torch.zeros_like(h), torch.zeros_like(c)


class LSTMTest(TensorTestCase):
    def setUp(self) -> None:
        torch.manual_seed(42)

    def assertLSTMOutputsEqual(
        self,
        actual_lstm: torch.nn.Module,
        reference_lstm: torch.nn.Module,
        inputs: torch.Tensor,
        state: Optional[tuple[torch.Tensor, torch.Tensor]] = None,
    ) -> None:
        actual_outputs, (actual_h, actual_c) = actual_lstm(inputs, state)
        target_outputs, (target_h, target_c) = reference_lstm(inputs, state)

        self.assertTensorEqual(target_outputs, actual_outputs)
        self.assertTensorEqual(target_h, actual_h)
        self.assertTensorEqual(target_c, actual_c)

    def test_lstm_equals_pytorch_lstm_on_batched_inputs(self) -> None:
        lstm, reference_lstm = create_lstm(
            input_size=2, hidden_size=3, bias=True, batch_first=False
        )
        inputs = torch.randn((4, 8, 2))
        self.assertLSTMOutputsEqual(lstm, reference_lstm, inputs)

    def test_lstm_batch_first_equals_pytorch_lstm(self) -> None:
        lstm, reference_lstm = create_lstm(
            input_size=2, hidden_size=3, bias=True, batch_first=True
        )
        inputs = torch.randn((8, 4, 2))
        self.assertLSTMOutputsEqual(lstm, reference_lstm, inputs)

    def test_lstm_equals_pytorch_lstm_without_batches(self) -> None:
        lstm, reference_lstm = create_lstm(
            input_size=2, hidden_size=3, bias=True, batch_first=True
        )
        inputs = torch.randn((4, 2))
        self.assertLSTMOutputsEqual(lstm, reference_lstm, inputs)

    def test_lstm_explicit_pass_state(self) -> None:
        lstm, reference_lstm = create_lstm(
            input_size=2, hidden_size=3, bias=True, batch_first=True
        )
        inputs = torch.randn((8, 4, 2))
        state = (torch.randn(1, 8, 3), torch.randn(1, 8, 3))
        self.assertLSTMOutputsEqual(lstm, reference_lstm, inputs, state)

    def test_lstm_uses_given_lstm_cell(self) -> None:
        lstm = LSTM(
            input_size=2,
            hidden_size=3,
            bias=True,
            batch_first=True,
            lstm_cell_factory=OutputsZeroLSTMCell,
        )

        inputs = torch.randn((8, 4, 2))
        expected_output = torch.zeros((8, 4, 3))
        expected_h, expected_c = torch.zeros((1, 8, 3)), torch.zeros((1, 8, 3))
        actual_output, (actual_h, actual_c) = lstm(inputs)

        self.assertTensorEqual(expected_output, actual_output)
        self.assertTensorEqual(expected_h, actual_h)
        self.assertTensorEqual(expected_c, actual_c)



import torch

from elasticai.creator.nn.linear import FixedPointLinear
from elasticai.creator.vhdl.number_representations import FixedPointFactory
from elasticai.creator.vhdl.translator.abstract.layers import FPLinear1dModule


def build_fp_linear_1d(
    layer: FixedPointLinear,
    layer_id: str,
    work_library_name: str,
) -> FPLinear1dModule:
    def to_list(tensor: torch.Tensor) -> list:
        return tensor.detach().numpy().tolist()

    return FPLinear1dModule(
        layer_id=layer_id,
        weight=to_list(layer.weight),
        bias=to_list(layer.bias),
        fixed_point_factory=layer.fixed_point_factory,
        work_library_name=work_library_name,
    )



from typing import Protocol

import torch

from elasticai.creator.nn.autograd_functions.fixed_point_quantization import (
    FixedPointDequantFunction,
    FixedPointQuantFunction,
)
from elasticai.creator.vhdl.number_representations import FixedPointFactory


class Arithmetics(Protocol):
    def quantize(self, a: torch.Tensor) -> torch.Tensor:
        ...

    def clamp(self, a: torch.Tensor) -> torch.Tensor:
        ...

    def round(self, a: torch.Tensor) -> torch.Tensor:
        ...

    def add(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        ...

    def sum(self, tensor: torch.Tensor, *tensors: torch.Tensor) -> torch.Tensor:
        ...

    def mul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        ...

    def matmul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        ...


class FloatArithmetics(Arithmetics):
    def quantize(self, a: torch.Tensor) -> torch.Tensor:
        return a

    def clamp(self, a: torch.Tensor) -> torch.Tensor:
        return a

    def round(self, a: torch.Tensor) -> torch.Tensor:
        return a

    def add(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return a + b

    def sum(self, tensor: torch.Tensor, *tensors: torch.Tensor) -> torch.Tensor:
        summed = tensor
        for t in tensors:
            summed += t
        return summed

    def mul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return a * b

    def matmul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return torch.matmul(a, b)


class FixedPointArithmetics(Arithmetics):
    def __init__(self, fixed_point_factory: FixedPointFactory) -> None:
        self.fixed_point_factory = fixed_point_factory

    def quantize(self, a: torch.Tensor) -> torch.Tensor:
        return self.round(self.clamp(a))

    def clamp(self, a: torch.Tensor) -> torch.Tensor:
        total_bits, frac_bits = (
            self.fixed_point_factory.total_bits,
            self.fixed_point_factory.frac_bits,
        )
        min_fp = -1 * (1 << (total_bits - 1)) / (1 << frac_bits)
        max_fp = int("1" * (total_bits - 1), 2) / (1 << frac_bits)
        return torch.clamp(a, min=min_fp, max=max_fp)

    def round(self, a: torch.Tensor) -> torch.Tensor:
        def float_to_int(x: torch.Tensor) -> torch.Tensor:
            return FixedPointQuantFunction.apply(x, self.fixed_point_factory)

        def int_to_fixed_point(x: torch.Tensor) -> torch.Tensor:
            return FixedPointDequantFunction.apply(x, self.fixed_point_factory)

        return int_to_fixed_point(float_to_int(a))

    def add(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.clamp(a + b)

    def sum(self, tensor: torch.Tensor, *tensors: torch.Tensor) -> torch.Tensor:
        summed = tensor
        for t in tensors:
            summed += t
        return self.clamp(summed)

    def mul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.round(self.clamp(a * b))

    def matmul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.round(self.clamp(torch.matmul(a, b)))



import torch

from elasticai.creator.nn.arithmetics import FixedPointArithmetics, FloatArithmetics
from elasticai.creator.tests.unit.tensor_test_case import TensorTestCase
from elasticai.creator.vhdl.number_representations import FixedPoint


class FloatArithmeticsTest(TensorTestCase):
    def setUp(self) -> None:
        self.ops = FloatArithmetics()

    def test_quantize(self) -> None:
        a = torch.tensor([0, 1.5, 2.25, 3.123456])
        actual = self.ops.quantize(a)
        self.assertTensorEqual(a, actual)

    def test_round(self) -> None:
        a = torch.tensor([0, 1.5, 2.25, 3.123456])
        actual = self.ops.round(a)
        self.assertTensorEqual(a, actual)

    def test_clamp(self) -> None:
        a = torch.tensor([0, 1.5, 2.25, 3.123456])
        actual = self.ops.clamp(a)
        self.assertTensorEqual(a, actual)

    def test_add(self) -> None:
        a = torch.tensor([1, 2, 3])
        b = torch.tensor([4, 5, 6])
        actual = self.ops.add(a, b)
        expected = [5, 7, 9]
        self.assertTensorEqual(expected, actual)

    def test_sum(self) -> None:
        a = torch.tensor([1, 2, 3])
        b = torch.tensor([-4, 5, -6])
        c = torch.tensor([1, 1, 1])
        actual = self.ops.sum(a, b, c)
        expected = [-2, 8, -2]
        self.assertTensorEqual(expected, actual)

    def test_mul(self) -> None:
        a = torch.tensor([1, 2, 3])
        b = torch.tensor([4, 5, 6])
        actual = self.ops.mul(a, b)
        expected = [4, 10, 18]
        self.assertTensorEqual(expected, actual)

    def test_matmul(self) -> None:
        a = torch.tensor([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        b = torch.tensor([[1], [2], [3]])
        actual = self.ops.matmul(a, b)
        expected = [[14], [32], [50]]
        self.assertTensorEqual(expected, actual)


class FixedPointArithmeticsTest(TensorTestCase):
    def setUp(self) -> None:
        fp_factory = FixedPoint.get_factory(total_bits=4, frac_bits=2)
        self.total_bits = fp_factory.total_bits
        self.frac_bits = fp_factory.frac_bits
        self.min_fp = -1 * (1 << (self.total_bits - 1)) / (1 << self.frac_bits)
        self.max_fp = int("1" * (self.total_bits - 1), 2) / (1 << self.frac_bits)
        self.ops = FixedPointArithmetics(fixed_point_factory=fp_factory)

    def test_quantize_clamps_minus5_to_minus2(self) -> None:
        a = torch.tensor([-5.0])
        actual = self.ops.quantize(a)
        expected = [-2.0]
        self.assertTensorEqual(expected, actual)

    def test_quantize_rounds_1_1_to_1_0(self) -> None:
        a = torch.tensor([1.1])
        actual = self.ops.quantize(a)
        expected = [1.0]
        self.assertTensorEqual(expected, actual)

    def test_clamp(self) -> None:
        a = torch.tensor([-5, -2.1, -1.0, 0.0, 1.8, 5])
        actual = self.ops.clamp(a)
        expected = [self.min_fp, self.min_fp, -1.0, 0.0, self.max_fp, self.max_fp]
        self.assertTensorEqual(expected, actual)

    def test_round(self) -> None:
        a = torch.tensor([-2.0, -1.8, 1.25, 1.6])
        actual = self.ops.round(a)
        expected = [-2.0, -1.75, 1.25, 1.5]
        self.assertTensorEqual(expected, actual)

    def test_add(self) -> None:
        a = torch.tensor([-0.25, 0.5, 1.0])
        b = torch.tensor([-1.5, 1.0, 1.5])
        actual = self.ops.add(a, b)
        expected = [-1.75, 1.5, 1.75]
        self.assertTensorEqual(expected, actual)

    def test_sum(self) -> None:
        a = torch.tensor([-0.25, 0.5, 1.0])
        b = torch.tensor([-1.5, 1.0, 1.5])
        c = torch.tensor([-0.5, -1.0, -1.0])
        actual = self.ops.sum(a, b, c)
        expected = [-2.0, 0.5, 1.5]
        self.assertTensorEqual(expected, actual)

    def test_mul(self) -> None:
        a = torch.tensor([-0.5, 1.5, 0.5])
        b = torch.tensor([0.5, 1.5, 1.25])
        actual = self.ops.mul(a, b)
        expected = [-0.25, 1.75, 0.5]
        self.assertTensorEqual(expected, actual)

    def test_matmul(self) -> None:
        a = torch.tensor([[-2.0, -1.75, -1.5], [-0.25, 0.0, 0.25], [1.25, 1.5, 1.75]])
        b = torch.tensor([[-0.25], [0.5], [0.25]])
        actual = self.ops.matmul(a, b)
        expected = [[-0.75], [0.0], [0.75]]
        self.assertTensorEqual(expected, actual)



import torch

from elasticai.creator.nn.linear import FixedPointLinear
from elasticai.creator.nn.lstm import FixedPointLSTMWithHardActivations
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.translator.pytorch import translator


class Model(torch.nn.Module):
    def __init__(self) -> None:
        super().__init__()

        fp_factory = FixedPoint.get_factory(8, 4)

        self.lstm = FixedPointLSTMWithHardActivations(
            input_size=3,
            hidden_size=5,
            bias=True,
            batch_first=True,
            fixed_point_factory=FixedPoint.get_factory(8, 4),
        )

        self.linear = FixedPointLinear(
            in_features=5, out_features=1, bias=True, fixed_point_factory=fp_factory
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.linear(self.lstm(x))


def main() -> None:
    model = Model()
    translation_args = {
        "0": {"work_library_name": "work"},
        "1": {"work_library_name": "work"},
    }
    code_repr = translator.translate_model(model, translation_args=translation_args)
    code_modules = list(code_repr)
    translator.save_code(code_modules, "build/")


if __name__ == "__main__":
    main()



from contextlib import contextmanager
from importlib import resources
from pathlib import PurePath
from typing import AnyStr, Iterable, TextIO

PathType = str | PurePath
Package = resources.Package


@contextmanager
def get_file(package: Package, file_name: str) -> TextIO:
    file_not_found = True
    for resource in resources.files(package).iterdir():
        if resource.name == file_name:
            file_not_found = False
            with resources.as_file(resource) as file:
                with open(file, "r") as opened_file:
                    yield opened_file
    if file_not_found:
        raise FileNotFoundError(
            f"The file '{file_name}' in package '{package}' does not exist."
        )


def read_text(
    package: Package, file_name: str, encoding: str = "utf-8"
) -> Iterable[AnyStr]:
    with get_file(package, file_name) as file:
        yield from map(lambda line: line.rstrip("\n"), file)


def read_bytes(package: Package, file_name: str) -> Iterable[bytes]:
    with get_file(package, file_name) as file:
        yield from file


def copy_file(package: Package, file_name: str, destination: PathType) -> None:
    data = read_bytes(package, file_name)
    with open(destination, "wb") as out_file:
        out_file.write(data)


def get_full_path(package: Package, file_name: str) -> str:
    resource = get_file(package, file_name)
    with resources.as_file(resource) as file:
        return file.as_posix()


def read_text_from_path(path: PathType, encoding: str = "utf-8") -> str:
    with open(path, "r", encoding=encoding) as inp_file:
        return inp_file.read()


def save_text_to_path(text: str, path: PathType, encoding: str = "utf-8") -> None:
    with open(path, "w", encoding=encoding) as out_file:
        out_file.write(text)



import unittest

from elasticai.creator.vhdl.hw_equivalent_layers.layers import AbstractTranslatableLayer


class MyRootLayer(AbstractTranslatableLayer):
    pass


class TranslatableRootLayerTest(unittest.TestCase):
    """
    Use AbstractTranslatableLayer to inject a different tracer and return a simplified graph representation.
    I don't think we need to inherit from torch.nn.Module in AbstractTranslatableLayer then.
    """

    ...



import unittest

from elasticai.creator.vhdl.code_files.fp_hard_sigmoid_file import FPHardSigmoidFile
from elasticai.creator.vhdl.number_representations import FixedPoint


class FPHardSigmoidComponentTest(unittest.TestCase):
    def test_hard_sigmoid_correct_number_of_lines(self) -> None:
        to_fp = FixedPoint.get_factory(total_bits=8, frac_bits=4)

        component = FPHardSigmoidFile(
            layer_id="0",
            zero_threshold=to_fp(-3),
            one_threshold=to_fp(3),
            slope=to_fp(0.15),
            y_intercept=to_fp(0.5),
            fixed_point_factory=to_fp,
        )

        self.assertEqual(len(list(component.code())), 88)



from abc import abstractmethod
from enum import Enum, auto
from itertools import chain
from typing import Any, Iterable, Protocol

from elasticai.creator.vhdl.code import Code, Translatable


class HWBlockInterface(Protocol):
    @abstractmethod
    def signal_definitions(self, prefix: str) -> Code:
        ...

    @abstractmethod
    def instantiation(self, prefix: str) -> Code:
        ...


class TranslatableHWBlockInterface(HWBlockInterface, Translatable, Protocol):
    ...


class SignalEnum(str, Enum):
    def _generate_next_value_(
        name: str, start: int, count: int, last_values: list[Any]
    ) -> Any:
        return name.lower()

    def __str__(self):
        return self.value


class StdLogicInSignals(SignalEnum):
    CLOCK = auto()
    ENABLE = auto()


class LogicVectorInSignals(SignalEnum):
    X = auto()
    Y_ADDRESS = auto()


class LogicVectorOutSignals(SignalEnum):
    Y = auto()
    X_ADDRESS = auto()


class StdLogicOutSignals(SignalEnum):
    DONE = auto()
    CLOCK = auto()


SIGNAL_MAPPING = (
    (StdLogicOutSignals.CLOCK, StdLogicInSignals.CLOCK),
    (StdLogicOutSignals.DONE, StdLogicInSignals.ENABLE),
    (LogicVectorOutSignals.Y, LogicVectorInSignals.X),
    (LogicVectorOutSignals.X_ADDRESS, LogicVectorInSignals.Y_ADDRESS),
)


class BaseHWBlock(HWBlockInterface):
    _vector_signals: tuple[SignalEnum, ...] = (
        LogicVectorInSignals.X,
        LogicVectorOutSignals.Y,
    )
    _logic_signals: tuple[SignalEnum, ...] = (
        StdLogicInSignals.ENABLE,
        StdLogicInSignals.CLOCK,
    )

    def __init__(self, x_width: int, y_width: int):
        self._x_width = x_width
        self._y_width = y_width

    def _get_vector_signal_name_to_signal_width_mapping(
        self,
    ) -> Iterable[tuple[str, int]]:
        return map(
            lambda s: (str(s), getattr(self, f"_{s}_width")),
            self._vector_signals,
        )

    def _get_logic_signal_names(self) -> Iterable[str]:
        return map(str, self._logic_signals)

    def signal_definitions(self, prefix: str) -> Code:
        yield from generate_signal_definitions(
            prefix,
            std_logic_signals=self._get_logic_signal_names(),
            logic_vector_signals=self._get_vector_signal_name_to_signal_width_mapping(),
        )

    def instantiation(self, prefix: str) -> Code:
        yield from instantiate_component(
            prefix=prefix, signals=chain(self._vector_signals, self._logic_signals)
        )


class BufferedBaseHWBlock(BaseHWBlock):
    _vector_signals = (
        LogicVectorInSignals.X,
        LogicVectorOutSignals.Y,
        LogicVectorOutSignals.X_ADDRESS,
        LogicVectorInSignals.Y_ADDRESS,
    )
    _logic_signals = (
        StdLogicInSignals.ENABLE,
        StdLogicInSignals.CLOCK,
        StdLogicOutSignals.DONE,
    )

    def __init__(
        self,
        x_width: int,
        y_width: int,
        x_address_width: int,
        y_address_width: int,
    ):
        super().__init__(y_width=y_width, x_width=x_width)
        self._x_address_width = x_address_width
        self._y_address_width = y_address_width


def generate_signal_definitions(
    prefix: str,
    std_logic_signals: Iterable[str],
    logic_vector_signals: Iterable[tuple[str, int]],
) -> Code:
    def _logic_vector(suffix, width) -> str:
        return f"signal {prefix}_{suffix} : std_logic_vector({width - 1} downto 0);"

    def _std_logic(suffix) -> str:
        return f"signal {prefix}_{suffix} : std_logic := '0';"

    code = [_std_logic(suffix) for suffix in std_logic_signals]
    code.extend(
        [_logic_vector(suffix, width) for suffix, width in logic_vector_signals]
    )

    return code


def instantiate_component(prefix: str, signals: Iterable[SignalEnum]) -> Code:
    def _generate_port_to_signal_connection(port: SignalEnum) -> str:
        return f"{port} => {prefix}_{port},"

    def _remove_comma_from_last_signal(signals: list[str]) -> list[str]:
        signals[-1] = signals[-1][:-1]
        return signals

    def _generate_port_map_signals(signals) -> list[str]:
        return [_generate_port_to_signal_connection(port) for port in signals]

    name = prefix
    code = [
        f"{name} : entity work.{name}(rtl)",
        "port map(",
    ]
    mapping = _generate_port_map_signals(signals)
    mapping = _remove_comma_from_last_signal(mapping)
    code.extend(mapping)
    code.append(");")
    return code



from elasticai.creator.tests.unit.vhdl.vhdl_file_testcase import GeneratedVHDLCodeTest
from elasticai.creator.vhdl.code_files.rom_component import RomFile
from elasticai.creator.vhdl.number_representations import float_values_to_fixed_point


class GenerateROMVhdTest(GeneratedVHDLCodeTest):
    def test_compare_files(self) -> None:
        # biases for the input gate
        bi = [1.1, 2.2, 3.3, 4.4, 5.5, 6.6]

        rom = RomFile(
            rom_name="rom_bi",
            layer_id="0",
            values=float_values_to_fixed_point(bi, total_bits=12, frac_bits=4),
            resource_option="auto",
        )
        generated_code = list(rom.code())

        expected_code = [
            "library ieee;",
            "    use ieee.std_logic_1164.all;",
            "    use ieee.std_logic_unsigned.all;",
            "entity rom_bi_0 is",
            "    port (",
            "        clk : in std_logic;",
            "        en : in std_logic;",
            "        addr : in std_logic_vector(3-1 downto 0);",
            "        data : out std_logic_vector(12-1 downto 0)",
            "    );",
            "end entity rom_bi_0;",
            "architecture rtl of rom_bi_0 is",
            (
                "    type rom_bi_0_array_t is array (0 to 2**3-1) of"
                " std_logic_vector(12-1 downto 0);"
            ),
            (
                "    signal ROM :"
                ' rom_bi_0_array_t:=(x"011",x"023",x"034",x"046",x"058",x"069",x"000",x"000");'
            ),
            "    attribute rom_style : string;",
            '    attribute rom_style of ROM : signal is "auto";',
            "begin",
            "    ROM_process: process(clk)",
            "    begin",
            "        if rising_edge(clk) then",
            "            if (en = '1') then",
            "                data <= ROM(conv_integer(addr));",
            "            end if;",
            "        end if;",
            "    end process ROM_process;",
            "end architecture rtl;",
        ]

        self.assertEqual(expected_code, generated_code)



from typing import Any

import torch

from elasticai.creator.vhdl.number_representations import FixedPointFactory


class FixedPointQuantFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx: Any, *args: Any, **kwargs: Any) -> torch.Tensor:
        if len(args) != 2:
            raise TypeError(
                "apply() takes exactly two arguments "
                "(x: torch.Tensor, fixed_point_factory: Callable[[float], FixedPoint])"
            )
        x: torch.Tensor = args[0]
        fp_factory: FixedPointFactory = args[1]
        total_bits, frac_bits = fp_factory.total_bits, fp_factory.frac_bits

        fp_ints = (x * (1 << frac_bits)).int().float()

        min_fp_int = 2 ** (total_bits - 1) * (-1)
        max_fp_int = 2 ** (total_bits - 1) - 1
        out_of_bounds = fp_ints[(fp_ints < min_fp_int) | (fp_ints > max_fp_int)]

        if torch.any(out_of_bounds):
            raise ValueError("Cannot quantize tensor. Values out of bounds.")

        return fp_ints

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        return *grad_outputs, None


class FixedPointDequantFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx: Any, *args: Any, **kwargs: Any) -> torch.Tensor:
        if len(args) != 2:
            raise TypeError(
                "apply() takes exactly two arguments "
                "(x: torch.Tensor, fixed_point_factory: Callable[[float], FixedPoint])"
            )
        x: torch.Tensor = args[0]
        fp_factory: FixedPointFactory = args[1]
        total_bits, frac_bits = fp_factory.total_bits, fp_factory.frac_bits

        fp_values = x / (1 << frac_bits)

        min_fp = -1 * (1 << (total_bits - 1)) / (1 << frac_bits)
        max_fp = int("1" * (total_bits - 1), 2) / (1 << frac_bits)
        out_of_bounds = fp_values[(fp_values < min_fp) | (fp_values > max_fp)]

        if torch.any(out_of_bounds):
            raise ValueError("Cannot dequantize tensor. Values out of bounds.")

        return fp_values

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        return *grad_outputs, None



from collections.abc import Collection
from dataclasses import dataclass

from elasticai.creator.vhdl.code import CodeFile, CodeModuleBase
from elasticai.creator.vhdl.code_files.fp_hard_tanh_component import FPHardTanhComponent
from elasticai.creator.vhdl.number_representations import FixedPointFactory


@dataclass
class FPHardTanhModule(CodeModuleBase):
    fixed_point_factory: FixedPointFactory
    layer_id: str

    @property
    def name(self) -> str:
        return self.layer_id

    @property
    def files(self) -> Collection[CodeFile]:
        yield FPHardTanhComponent(
            min_val=self.fixed_point_factory(-1),
            max_val=self.fixed_point_factory(1),
            fixed_point_factory=self.fixed_point_factory,
            layer_id=self.layer_id,
        )



from itertools import chain, repeat
from string import Template
from typing import Callable, Iterable, Iterator, Union

from elasticai.creator.resource_utils import read_text
from elasticai.creator.vhdl.code import Code, TemplateCodeFile


class VHDLFile(TemplateCodeFile):
    """
    `VHDLFile` helps you to fill templates with values.
    It loads a vhdl template file from the package `elasticai.creator.vhdl.template` and fills template parameters
    with the parameters passed to the constructor.
    Each of these parameters is expected to be either a `str` or a `Iterable[str]`.
    Depending on the type, the parameter will be subject to multi or single line expansion.
    Multi line expansion in the context means the template `"$myValue"` with the parameter `{"myValue": ["1", "2"]}`
    will result in the string `"1\n2"`.
    The **code** method returns the lines of code resulting from the parameters filled into the template.
    """

    _template_package = "elasticai.creator.vhdl.templates"

    def __init__(self, name: str, **parameters: Union[str, Iterable[str]]) -> None:
        self._name = name
        (
            self._parameters,
            self._multiline_parameters,
        ) = self._split_single_and_multiline_parameters(parameters)

    @staticmethod
    def _split_single_and_multiline_parameters(parameters: dict):
        single_line_parameters = dict(
            filter(lambda i: isinstance(i[1], str), parameters.items())
        )
        multiline_parameters = dict(
            filter(lambda i: not isinstance(i[1], str), parameters.items())
        )
        return single_line_parameters, multiline_parameters

    @property
    def single_line_parameters(self) -> dict[str, str]:
        return dict(**self._parameters)

    @property
    def parameters(self):
        return dict(chain(self._parameters.items(), self._multiline_parameters.items()))

    @property
    def multi_line_parameters(self):
        return dict(**self._multiline_parameters)

    @property
    def name(self) -> str:
        return f"{self._name}.vhd"

    def code(self) -> Code:
        template = read_text(self._template_package, f"{self._name}.tpl.vhd")
        template = expand_template(template, **self._parameters)
        template = expand_multiline_template(template, **self._multiline_parameters)
        return template


def expand_multiline_template(
    template: Union[str, Iterable[str]], **kwargs: Iterable[str]
) -> Iterator[str]:
    """Expand a template field to multiple lines, while keeping indentation.
    Example:
        >>> template = "\\t$my_key"
        >>> values = ["hello,", "world", "!"]
        >>> "\\n".join(expand_multiline_template(template, my_key=values))
        '\\thello,\\n\\tworld\\n\\t!'
    """
    lines = _unify_template_datatype(template)
    for line in lines:
        contains_no_key = True
        for key in kwargs:
            if f"${key}" in line:
                contains_no_key = False
                for placeholder_line, value in zip(repeat(line), kwargs[key]):
                    t = Template(placeholder_line)
                    yield t.safe_substitute({key: value})
                break
        if contains_no_key:
            yield line


def _unify_template_datatype(template: Union[str, Iterable[str]]) -> Iterable[str]:
    if hasattr(template, "splitlines") and callable(template.splitlines):
        lines = template.splitlines()
    else:
        lines = template
    return lines


def expand_template(template: str | Iterable[str], **kwargs: str) -> Iterator[str]:
    if isinstance(template, str):
        yield Template(template).safe_substitute(kwargs)
    else:
        for line in template:
            yield Template(line).safe_substitute(kwargs)



from elasticai.creator.vhdl.translator.build_function_mapping import (
    BuildFunctionMapping,
)
from elasticai.creator.vhdl.translator.pytorch.build_functions import (
    build_fixed_point_lstm,
    build_fp_hard_sigmoid,
    build_fp_hard_tanh,
    build_fp_linear_1d,
    build_fp_relu,
)

DEFAULT_BUILD_FUNCTION_MAPPING = BuildFunctionMapping(
    mapping={
        "elasticai.creator.nn.lstm.FixedPointLSTMWithHardActivations": (
            build_fixed_point_lstm
        ),
        "elasticai.creator.nn.linear.FixedPointLinear": build_fp_linear_1d,
        "elasticai.creator.nn.hard_sigmoid.HardSigmoid": build_fp_hard_sigmoid,
        "elasticai.creator.nn.relu.ReLU": build_fp_relu,
        "elasticai.creator.nn.hard_tanh.HardTanh": build_fp_hard_tanh,
    }
)



import unittest

from elasticai.creator.vhdl.code_files.fp_hard_tanh_component import FPHardTanhComponent
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.translator.abstract.layers.fp_hard_tanh_module import (
    FPHardTanhModule,
)


class FPHardTanhModuleTest(unittest.TestCase):
    def setUp(self) -> None:
        self.fp_factory = FixedPoint.get_factory(total_bits=8, frac_bits=4)
        self.module = FPHardTanhModule(
            layer_id="0", fixed_point_factory=self.fp_factory
        )

    def test_components_return_only_one_component(self) -> None:
        components = list(self.module.files)
        self.assertEqual(len(components), 1)
        self.assertEqual(type(components[0]), FPHardTanhComponent)

    def test_components_component_args_are_correctly_set(self) -> None:
        component = list(self.module.files)[0]
        self.assertEqual(component.single_line_parameters["min_val"], str(self.fp_factory(-1).to_signed_int()))  # type: ignore
        self.assertEqual(component.single_line_parameters["max_val"], str(self.fp_factory(1).to_signed_int()))  # type: ignore



import unittest

from elasticai.creator.tests.text_parsing import extract_section


class ExtractSectionTest(unittest.TestCase):
    def test_extract_CD_from_AACBAADB(self):
        text = [f"{c}" for c in "AACBAADB"]
        self.assertEqual(
            [["C"], ["D"]],
            extract_section(begin=["A", "A"], end="B", lines=text),
        )

    def test_extract_CD_from_AACBBAADBB(self):
        text = [f"{c}" for c in "AACBBAADBB"]
        self.assertEqual(
            [["C"], ["D"]],
            extract_section(begin=["A", "A"], end=["B", "B"], lines=text),
        )

    def test_extract_CD_from_ABCAABDAA(self):
        text = [f"{c}" for c in "ABCAABDAA"]
        self.assertEqual(
            [["C"], ["D"]],
            extract_section(begin=["A", "B"], end=["A", "A"], lines=text),
        )



import unittest

from elasticai.creator.vhdl.code_files.fp_linear_1d_component import FPLinear1dFile
from elasticai.creator.vhdl.code_files.rom_component import RomFile
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.translator.abstract.layers import FPLinear1dModule


class FPLinear1dModuleTest(unittest.TestCase):
    def test_contains_all_needed_components(self) -> None:
        linear = FPLinear1dModule(
            layer_id="ll1",
            weight=[[1.0, 2.0, 3.0, 4.0, 5.0, 6.0]],
            bias=[1.0, 2.0],
            fixed_point_factory=FixedPoint.get_factory(total_bits=8, frac_bits=4),
        )
        vhdl_components = linear.files

        target_components = [
            (
                FPLinear1dFile,
                "fp_linear_1d_{layer_name}.vhd".format(layer_name=linear.layer_id),
            ),
            (
                RomFile,
                "w_rom_fp_linear_1d_{layer_name}.vhd".format(
                    layer_name=linear.layer_id
                ),
            ),
            (
                RomFile,
                "b_rom_fp_linear_1d_{layer_name}.vhd".format(
                    layer_name=linear.layer_id
                ),
            ),
        ]
        actual_components = [(type(x), x.name) for x in vhdl_components]

        self.assertEqual(target_components, actual_components)



import typing
from abc import ABC, abstractmethod
from typing import Any

import torch.nn

from elasticai.creator.mlframework import Module
from elasticai.creator.nn.hard_sigmoid import HardSigmoid as nnHardSigmoid
from elasticai.creator.nn.linear import FixedPointLinear as nnFixedPointLinear
from elasticai.creator.vhdl.code import Code, CodeModule, CodeModuleBase, Translatable
from elasticai.creator.vhdl.code_files.utils import calculate_address_width
from elasticai.creator.vhdl.hw_equivalent_layers.hw_blocks import (
    BaseHWBlock,
    BufferedBaseHWBlock,
    HWBlockInterface,
)
from elasticai.creator.vhdl.model_tracing import (
    HWEquivalentTracer,
    Tracer,
    create_hw_block_collection,
)
from elasticai.creator.vhdl.number_representations import FixedPointFactory
from elasticai.creator.vhdl.vhdl_files import VHDLFile


class AbstractTranslatableLayer(Translatable, HWBlockInterface, ABC):
    def __init__(self, hw_block: HWBlockInterface):
        self._hw_block = hw_block

    @abstractmethod
    def _get_tracer(self) -> Tracer:
        ...

    @abstractmethod
    def _template_parameters(self) -> dict[str, str]:
        return {
            name: str(getattr(self, name))
            for name in ("x_width", "y_width", "x_address_width", "y_address_width")
        }

    def signal_definitions(self, prefix: str) -> Code:
        return self._hw_block.signal_definitions(prefix)

    def instantiation(self, prefix: str) -> Code:
        return self._hw_block.instantiation(prefix)

    def translate(self) -> CodeModule:
        module: Module = typing.cast(Module, self)
        graph = self._get_tracer().trace(module)
        blocks = create_hw_block_collection(graph)
        signals = blocks.signals("")
        layer_instantiations = blocks.instantiations("")
        layer_connections = ["fp_linear_x <= x;"]
        return CodeModuleBase(
            name="network",
            files=(
                VHDLFile(
                    name="network",
                    signal_definitions=signals,
                    layer_instantiations=layer_instantiations,
                    layer_connections=layer_connections,
                    **self._template_parameters(),
                ),
            ),
        )


class RootModule(torch.nn.Module, AbstractTranslatableLayer):
    def __init__(self):
        self.elasticai_tags = {
            "x_address_width": 1,
            "y_address_width": 1,
            "x_width": 1,
            "y_width": 1,
        }
        torch.nn.Module.__init__(self)
        AbstractTranslatableLayer.__init__(
            self, BufferedBaseHWBlock(**self.elasticai_tags)
        )

    def _template_parameters(self) -> dict[str, str]:
        return dict(((key, str(value)) for key, value in self.elasticai_tags.items()))

    def translate(self):
        self._hw_block = BufferedBaseHWBlock(**self.elasticai_tags)
        return super().translate()

    def _get_tracer(self) -> Tracer:
        return HWEquivalentTracer()


class FixedPointHardSigmoid(nnHardSigmoid, HWBlockInterface):
    def signal_definitions(self, prefix: str) -> Code:
        return self._hw_block.signal_definitions(prefix)

    def instantiation(self, prefix: str) -> Code:
        return self._hw_block.instantiation(prefix)

    def __init__(
        self,
        in_place: bool = False,
        *,
        data_width,
    ):
        super().__init__(in_place)
        self._hw_block = BaseHWBlock(x_width=data_width, y_width=data_width)


class FixedPointLinear(nnFixedPointLinear, HWBlockInterface):
    def signal_definitions(self, prefix: str) -> Code:
        return self._hw_block.signal_definitions(prefix)

    def instantiation(self, prefix: str) -> Code:
        return self._hw_block.instantiation(prefix)

    def __init__(
        self,
        in_features: int,
        out_features: int,
        fixed_point_factory: FixedPointFactory,
        bias: bool = True,
        device: Any = None,
        *,
        data_width: int,
    ):
        super().__init__(
            in_features=in_features,
            out_features=out_features,
            fixed_point_factory=fixed_point_factory,
            bias=bias,
            device=device,
        )
        self._hw_block = BufferedBaseHWBlock(
            y_width=data_width,
            x_width=data_width,
            x_address_width=calculate_address_width(in_features),
            y_address_width=calculate_address_width(out_features),
        )



from typing import Any, Collection, Iterable, Iterator

from elasticai.creator.mlframework import Module
from elasticai.creator.mlframework.typing import Parameter
from elasticai.creator.vhdl.code import (
    Code,
    CodeFile,
    CodeModule,
    CodeModuleBase,
    Translatable,
)
from elasticai.creator.vhdl.hw_equivalent_layers.typing import HWEquivalentLayer
from elasticai.creator.vhdl.model_tracing import (
    HWEquivalentGraph,
    HWEquivalentNode,
    Node,
)


class DummyModule(HWEquivalentLayer):
    def signal_definitions(self, prefix: str) -> Code:
        return []

    def instantiation(self, prefix: str) -> Code:
        return []

    def translate(self) -> CodeModule:
        return CodeModuleBase(name="", files=[])

    @property
    def training(self) -> bool:
        return True

    def extra_repr(self) -> str:
        return ""

    def named_children(self) -> Iterable[tuple[str, "Module"]]:
        yield from []

    def parameters(self, recurse: bool = True) -> Iterator[Parameter]:
        yield from []

    def state_dict(self) -> dict[str, Any]:
        return {}

    def __call__(self, x: Any, *args: Any, **kwargs: Any) -> Any:
        return x


class DummyModuleNode(Node):
    @property
    def op(self) -> str:
        return "call_module"

    def __init__(self, name=""):
        self._name = name

    @property
    def name(self) -> str:
        return self._name


class DummyHWEquivalentNode(DummyModuleNode, HWEquivalentNode):
    def __init__(self, name=""):
        super().__init__(name)
        self._module = DummyModule()

    @property
    def hw_equivalent_layer(self) -> HWEquivalentLayer:
        return self._module

    @hw_equivalent_layer.setter
    def hw_equivalent_layer(self, value):
        self._module = value


class DummyGraph(HWEquivalentGraph):
    @property
    def hw_equivalent_nodes(self) -> Iterable[HWEquivalentNode]:
        yield from []

    def __init__(self, nodes):
        self._nodes = nodes

    @property
    def nodes(self) -> Iterable[HWEquivalentNode]:
        return self._nodes


class DummyCodeModule(CodeModule):
    @property
    def files(self) -> Collection[CodeFile]:
        return []

    @property
    def submodules(self) -> Collection["CodeModule"]:
        return []

    @property
    def name(self) -> str:
        return ""


class TranslatableDummyModule(DummyModule, Translatable):
    def translate(self) -> "CodeModule":
        return DummyCodeModule()



from elasticai.creator.vhdl.number_representations import FixedPoint, FixedPointFactory
from elasticai.creator.vhdl.vhdl_files import VHDLFile


class FPHardSigmoidFile(VHDLFile):
    def __init__(
        self,
        layer_id: str,
        zero_threshold: FixedPoint,
        one_threshold: FixedPoint,
        slope: FixedPoint,
        y_intercept: FixedPoint,
        fixed_point_factory: FixedPointFactory,
    ):
        d = dict(
            data_width=fixed_point_factory.total_bits,
            frac_width=fixed_point_factory.frac_bits,
            one=fixed_point_factory(1).to_signed_int(),
            zero_threshold=zero_threshold.to_signed_int(),
            one_threshold=one_threshold.to_signed_int(),
            y_intercept=y_intercept.to_signed_int(),
            slope=slope.to_signed_int(),
            layer_name=layer_id,
        )
        stringified_d = dict(((k, str(v)) for k, v in d.items()))
        name = "fp_hard_sigmoid"
        super().__init__(name=name, **stringified_d)



import unittest

import torch

from elasticai.creator.nn.linear import FixedPointLinear
from elasticai.creator.vhdl.number_representations import FixedPoint
from elasticai.creator.vhdl.translator.pytorch.build_functions import build_fp_linear_1d


def aragnge_parameter(
    start: int, end: int, shape: tuple[int, ...]
) -> torch.nn.Parameter:
    return torch.nn.Parameter(
        torch.reshape(torch.arange(start, end, dtype=torch.float32), shape)
    )


class FPLinear1dBuildFunctionTest(unittest.TestCase):
    def test_weights_and_bias_correct_set(self) -> None:
        fp_factory = FixedPoint.get_factory(total_bits=8, frac_bits=4)

        linear = FixedPointLinear(
            fixed_point_factory=fp_factory, in_features=3, out_features=2, bias=True
        )
        linear.weight = aragnge_parameter(start=1, end=4, shape=(1, -1))
        linear.bias = aragnge_parameter(start=1, end=2, shape=(-1,))

        fp_linear1d = build_fp_linear_1d(
            linear, layer_id="ll1", work_library_name="work"
        )
        self.assertEqual(fp_linear1d.weight, [[1.0, 2.0, 3.0]])
        self.assertEqual(fp_linear1d.bias, [1.0])



from typing import Iterator, TextIO


class VHDLFileReaderWithoutComments:
    """
    Allows you to iterate over a text ignoring blank lines and vhdl comments.
    This is mainly used for testing. That way we can compare expected and actually generated
    code without considering formatting and comments
    """

    def __init__(self, file: TextIO):
        self._file = file

    @staticmethod
    def _line_is_relevant(line: str) -> bool:
        return len(line) > 0 and not line.startswith("--")

    @staticmethod
    def _strip_trailing_comment(line: str) -> str:
        return line.split(" --")[0]

    def as_list(self) -> list[str]:
        return list(self)

    def __iter__(self) -> Iterator[str]:
        for line in self._file:
            line = line.rstrip("\n")
            line = line.strip()
            if self._line_is_relevant(line):
                line = self._strip_trailing_comment(line)
                yield line
